# pyre-ignore-all-errors
# TODO: this file should be bxgateway, since it depends on bxgateway opts

import json
import os
from dataclasses import dataclass
from typing import List, Optional, TYPE_CHECKING, Union, Dict, Any
from argparse import Namespace

from bxcommon.models.bdn_account_model_base import BdnAccountModelBase
from bxcommon.models.blockchain_network_model import BlockchainNetworkModel
from bxcommon.models.blockchain_peer_info import BlockchainPeerInfo
from bxcommon.models.node_model import NodeModel
from bxcommon.models.outbound_peer_model import OutboundPeerModel
from bxcommon.utils import model_loader, config
from bxutils import log_messages
from bxutils import logging
from bxutils.encoding.json_encoder import EnhancedJSONEncoder

logger = logging.get_logger(__name__)

if TYPE_CHECKING:
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.common_opts import CommonOpts


@dataclass
class CacheNetworkInfo:
    source_version: str
    relay_peers: List[OutboundPeerModel]
    blockchain_networks: Dict[int, BlockchainNetworkModel]
    blockchain_peers: List[BlockchainPeerInfo]
    node_model: NodeModel
    accounts: Dict[str, Union[BdnAccountModelBase, Any]]


def update_cache_file(
    opts: "CommonOpts",
    potential_relay_peers: Optional[List[OutboundPeerModel]] = None,
    blockchain_peers: Optional[List[BlockchainPeerInfo]] = None,
    accounts: Optional[Dict[str, Optional[BdnAccountModelBase]]] = None
) -> None:
    data = read(opts)
    node_model = model_loader.load_model(NodeModel, opts.__dict__)
    if data is None:
        data = CacheNetworkInfo(
            source_version=opts.source_version,
            relay_peers=potential_relay_peers,
            blockchain_networks=opts.blockchain_networks,
            blockchain_peers=blockchain_peers,
            node_model=node_model,
            accounts=accounts
        )
    else:
        data.blockchain_networks = opts.blockchain_networks
        data.node_model = node_model
        if potential_relay_peers is not None:
            data.relay_peers = potential_relay_peers
        if blockchain_peers is not None:
            data.blockchain_peers = blockchain_peers
        if accounts is not None:
            if data.accounts:
                data.accounts.update(accounts)
            else:
                data.accounts = accounts

    try:
        cookie_file_path = config.get_data_file(opts.cookie_file_path)
        os.makedirs(os.path.dirname(cookie_file_path), exist_ok=True)
        with open(cookie_file_path, "w") as cookie_file:
            json.dump(data, cookie_file, indent=4, cls=EnhancedJSONEncoder)
    # pylint: disable=broad-except
    except Exception as e:
        logger.error(
            "Failed when tried to write to cache file: {} with exception: {}",
            opts.cookie_file_path,
            e
        )


def read(opts: Union["CommonOpts", Namespace]) -> Optional[CacheNetworkInfo]:
    cache_file_info = None
    enable_node_cache = opts.__dict__.get("enable_node_cache", False)
    relative_path = opts.__dict__.get("cookie_file_path", None)
    if enable_node_cache and relative_path is not None:
        try:
            cookie_file_path = config.get_data_file(relative_path)
            if os.path.exists(cookie_file_path):
                with open(cookie_file_path, "r") as cookie_file:
                    cache_file_info = model_loader.load_model(CacheNetworkInfo, json.load(cookie_file))
                    if cache_file_info.source_version != opts.source_version:
                        os.remove(cookie_file_path)
            else:
                logger.warning(log_messages.READ_CACHE_FILE_WARNING, cookie_file_path)
        # pylint: disable=broad-except
        except Exception as e:
            logger.error(
                "Failed when tried to read from cache file: {} with exception: {}",
                relative_path,
                str(e)
            )
    return cache_file_info



from typing import Union, List

from bxcommon.messages.abstract_message import AbstractMessage
from bxcommon.utils import convert
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.messages.abstract_block_message import AbstractBlockMessage

from bxgateway.connections.eth.eth_base_connection_protocol import EthBaseConnectionProtocol
from bxgateway.messages.eth.protocol.block_bodies_eth_protocol_message import BlockBodiesEthProtocolMessage
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType
from bxgateway.messages.eth.protocol.get_block_bodies_eth_protocol_message import GetBlockBodiesEthProtocolMessage
from bxgateway.messages.eth.protocol.receipts_eth_protocol_message import ReceiptsEthProtocolMessage
from bxgateway.messages.eth.protocol.status_eth_protocol_message import StatusEthProtocolMessage
from bxgateway.messages.eth.protocol.status_eth_protocol_message_v63 import StatusEthProtocolMessageV63
from bxgateway.utils.eth.rlpx_cipher import RLPxCipher
from bxutils import logging

logger = logging.get_logger(__name__)


class EthRemoteConnectionProtocol(EthBaseConnectionProtocol):
    def __init__(self, connection, is_handshake_initiator, rlpx_cipher: RLPxCipher):
        super().__init__(connection, is_handshake_initiator, rlpx_cipher)

        connection.message_handlers.update({
            EthProtocolMessageType.STATUS: self.msg_status,
            EthProtocolMessageType.GET_BLOCK_BODIES: self.msg_get_block_bodies,
            EthProtocolMessageType.BLOCK_HEADERS: self.msg_proxy_response,
            EthProtocolMessageType.BLOCK_BODIES: self.msg_block_bodies,
            EthProtocolMessageType.NODE_DATA: self.msg_proxy_response,
            EthProtocolMessageType.RECEIPTS: self.msg_block_receipts
        })

    def msg_status(self, msg: Union[StatusEthProtocolMessage, StatusEthProtocolMessageV63]):
        super(EthRemoteConnectionProtocol, self).msg_status(msg)
        self.connection.on_connection_established()
        self.node.remote_blockchain_connection_established = True
        self.node.on_remote_blockchain_connection_ready(self.connection)

        self.connection.schedule_pings()

    def msg_block_bodies(self, msg: BlockBodiesEthProtocolMessage) -> None:
        self.node.log_received_remote_blocks(len(msg.get_block_bodies_bytes()))
        self.msg_proxy_response(msg)

    def msg_block_receipts(self, msg: ReceiptsEthProtocolMessage) -> None:
        self.node.log_received_remote_blocks(len(msg.get_receipts_bytes()))
        self.msg_proxy_response(msg)

    def msg_get_block_bodies(self, msg: GetBlockBodiesEthProtocolMessage) -> None:
        block_hashes = msg.get_block_hashes()
        logger.debug(
            "Received unexpected get block bodies message from remote "
            "blockchain node for {} blocks: {}. Replying with empty block bodies.",
            len(block_hashes),
            ", ".join(
                [convert.bytes_to_hex(block_hash.binary) for block_hash in block_hashes[:10]]
            )
        )
        empty_block_bodies_msg = BlockBodiesEthProtocolMessage(None, [])
        self.connection.enqueue_msg(empty_block_bodies_msg)

    def msg_block(self, msg: AbstractBlockMessage) -> None:
        raise NotImplemented

    def _build_get_blocks_message_for_block_confirmation(
        self, hashes: List[Sha256Hash]
    ) -> AbstractMessage:
        pass

    def _set_transaction_contents(
        self, tx_hash: Sha256Hash, tx_content: Union[memoryview, bytearray]
    ) -> None:
        pass




from typing import Callable, Dict, List
import pycond as pc
from bxutils import logging

logger = logging.get_logger(__name__)


pc.ops_use_symbolic_and_txt(allow_single_eq=True)


def get_validator(filter_string: str) -> Callable[[Dict], bool]:
    logger.trace("Getting validator for filters {}", filter_string)
    res = pc.qualify(filter_string.lower(), brkts="()", add_cached=True)
    return res


def get_keys(filter_string: str) -> List[str]:
    logger.trace("Getting keys for filters {}", filter_string)
    _, nfos = pc.parse_cond(filter_string.lower(), brkts="()", add_cached=True)
    return nfos["keys"]



from abc import ABC, abstractmethod
from enum import Enum
from sys import getsizeof
from typing import Union, Deque, Set, Any, List, Optional, NamedTuple

from psutil import Process
from pympler import asizeof
from pympler.asizeof import Asized

DEFAULT_DETAILED_MEMORY_BREAKDOWN_LIMIT = 50 * 1024

_process: Optional[Process] = None


class SpecialTuple(NamedTuple):
    size: int
    seen_ids: Set[int]


class ObjectSize:
    """
    Represents size of object with detailed breakdown by object field
    """

    def __init__(self, name=None, size=0, flat_size=0, references=None, is_actual_size=True) -> None:
        self.name = name
        self.size = size
        self.flat_size = flat_size
        self.references = references
        self.is_actual_size = is_actual_size


class SpecialMemoryProperties(ABC):
    """
    Meta class that should be implemented in all classes with special memory types
    """

    @abstractmethod
    def special_memory_size(self, ids: Optional[Set[int]] = None) -> SpecialTuple:
        if ids is None:
            ids = set()
        return SpecialTuple(size=0, seen_ids=ids)


# pylint: disable=global-statement
def get_app_memory_usage() -> int:
    """
    Provides total application memory usage in bytes
    :return: int
    """
    global _process
    if _process is None:
        _process = Process()

    process = _process
    assert process is not None
    return process.memory_info().rss


def get_object_size(obj, sizer=asizeof):
    # type: (object, Any) -> ObjectSize
    """
    Calculates total size memory consumed by objects and all objects that it references
    :param obj: object to calculate size
    :param sizer: sizer object
    :return: object representing memory usage breakdown for the object
    """
    obj_size = sizer.asized(obj)
    return _to_size_obj(obj_size, DEFAULT_DETAILED_MEMORY_BREAKDOWN_LIMIT)


def get_detailed_object_size(obj, detailed_if_greater_than=DEFAULT_DETAILED_MEMORY_BREAKDOWN_LIMIT, sizer=asizeof):
    # type: (object, int, Any) -> ObjectSize
    """
    Calculates total size memory consumed by objects and all objects that it references.
    Generates breakdown by reference.
    :param obj: object to calculate size
    :param sizer: sizer object
    :param detailed_if_greater_than: break down memory by fields of object if it is size is greater than provided value
    :return: object representing memory usage breakdown for the object
    """
    obj_size = sizer.asized(obj, detail=10)
    return _to_size_obj(obj_size, detailed_if_greater_than)


def _to_size_obj(obj_size: Asized, detailed_if_greater_than: int) -> ObjectSize:
    references = []

    if obj_size.refs is not None:
        for ref in obj_size.refs:
            size_obj = _to_size_obj(ref, detailed_if_greater_than)
            if size_obj.size > detailed_if_greater_than:
                references.append(size_obj)

    return ObjectSize(obj_size.name, obj_size.size, obj_size.flat, references)


def _get_special_size_helper(obj: Union[Deque, Set[Any], memoryview, List], ids: Set[int]) -> SpecialTuple:
    """
    Calculates total size memory consumed by special objects and all special objects that it references
    :param obj: object to calculate size
    :param ids: set of already seen ids
    :return: Tuple of the size in bytes of the special data structures of that object, and the integer keys associated
    with those objects and their special contents
    """
    default_size = getsizeof(0)
    total_size = 0
    if id(obj) not in ids:
        if isinstance(obj, memoryview):
            total_size += getsizeof(obj, default_size)
            if id(obj.obj) not in ids:
                total_size += obj.nbytes
                ids.add(id(obj.obj))
            ids.add(id(obj))
            return SpecialTuple(size=total_size, seen_ids=ids)
        # The size of deques and memoryviews are not properly recorded using get_object_size
        elif isinstance(obj, Deque):
            total_size += getsizeof(obj, default_size)
            for elem in list(obj):
                if isinstance(elem, (Deque, Set, memoryview, List)):
                    curr_size, ids = _get_special_size_helper(elem, ids)
                    total_size += curr_size
                else:
                    total_size += get_object_size(elem).size
                ids.add(id(elem))
        else:
            for elem in list(obj):
                if isinstance(elem, (Deque, Set, memoryview, List)):
                    curr_size, ids = _get_special_size_helper(elem, ids)
                    total_size += curr_size
                    ids.add(id(elem))
        ids.add(id(obj))
    return SpecialTuple(size=total_size, seen_ids=ids)


def get_special_size(obj: Any, ids: Optional[Set[int]] = None) -> SpecialTuple:
    """
    Handles size requests for classes and data structures
    :param obj: object to calculate size
    :param ids: set of already seen ids
    :return: Tuple of the size in bytes of the special data structures of that object, and the integer keys associated
    with those objects and their special contents
    """
    if ids is None:
        ids = set()
    if obj is None:
        # pyre-fixme[6]: Expected `int` for 1st param but got `None`.
        ids.add(obj)
        return SpecialTuple(size=0, seen_ids=ids)
    if isinstance(obj, SpecialMemoryProperties):
        return obj.special_memory_size(ids)
    elif isinstance(obj, (Deque, Set, memoryview, List)):
        return _get_special_size_helper(obj, ids)
    else:
        return SpecialTuple(size=0, seen_ids=ids)


def add_special_objects(*args, ids: Optional[Set[int]] = None) -> SpecialTuple:
    """
    Should be called by class methods with arguments consisting of attributes of the class that are objects
    with deques, memoryviews, etc. within them ...
    OR
    deques, memoryviews, etc. data structures themselves
    See AbstractConnection, InputBuffer, for examples
    :param args: attributes of a class to be used in the special size calculation
    :param ids: set of already seen ids
    :return: Tuple of the size in bytes of the special data structures of that object, and the integer keys associated
    with those objects and their special contents
    """
    total_size = 0
    if ids is None:
        ids = set()
    for obj in args:
        curr_size, ids = get_special_size(obj, ids=ids)
        total_size += curr_size
    return SpecialTuple(size=total_size, seen_ids=ids)


class ObjectType(Enum):

    BASE = "Base"
    META = "Meta"
    MAP_PROXY = "MapProxy"
    DEFAULT_MAP_PROXY = "DefaultMapProxy"
    MAIN_TASK_BASE = "MainTaskBase"
    TASK_QUEUE_PROXY = "TaskQueueProxy"

    def __str__(self) -> str:
        return str(self.value)


class SizeType(Enum):
    SPECIAL = "Special"
    OBJECT = "Object"
    TRUE = "True"
    ESTIMATE = "Estimate"

    def __str__(self) -> str:
        return str(self.value)



import struct

from bxutils import logging

logger = logging.get_logger(__name__)


class PayloadElement:
    """payload element

    name (str): element block name
    structure (str): python struct packing format

    optionally provide decode,encode functions to be applied
    """
    def __init__(self, name, structure, default=None, decode=None, encode=None) -> None:
        self.name = name
        self.structure = structure
        self.default = default
        self.size = struct.calcsize(self.structure)
        self.decode = decode or (lambda x: x)
        self.encode = encode or (lambda x: x)
        self.block_name = None
        self.block_version = None
        self.offset = None


class PayloadBlock:
    """PayloadBlock element use to create a message block template from PayloadElements and PayloadBlocks

        Args:
            off (int): block offset
            name (str): block name
            version (int): block version
            * (PayloadElement/PayloadBlock)
    """
    def __init__(self, off, name, version, *args) -> None:
        elements = []
        self.name = name
        self.version = version
        self.start_offset = off
        self.off = off
        self.size = 0
        self.structure = ""
        self.elements = self._iter_input(elements, args, name, version)

    def _iter_input(self, elements, iterator, name, version):
        for item in iterator:
            if isinstance(item, PayloadElement):
                elements.append(item)
                # add context metadata for block element
                item.block_name = name
                item.block_version = version
                item.offset = self.off

                self.off += item.size
                self.size += item.size
            elif isinstance(item, PayloadBlock):
                self._iter_input(elements, item.elements, name=item.name, version=item.version)
            else:
                raise Exception("unknown input")
        return elements

    def build(self, buf, **kwargs):
        """build message buffer from kwargs"""
        for element in self.elements:
            s = kwargs.get(element.name, element.default)
            if element.offset + element.size > len(buf) and s is not None:
                logger.trace("cannot pack {},{} for {}.{} buffer too small {} {} {}",
                             s, element.name, element.block_name, element.block_version, len(buf), self.name,
                             self.version)
                continue
            if not s:
                continue
            if element.encode:
                s = element.encode(s)
            try:
                struct.pack_into(element.structure, buf, element.offset, s)
            except Exception as e:
                logger.debug("{} cannot pack {} into {} for {} {} {} {}",
                             repr(e), [s], element.structure, element.name, element.offset, len(buf),
                             [w.name for w in self.elements])
                raise Exception
        return buf

    def read(self, buf):
        """unpacks buffer contents into dictionary"""
        contents = dict()
        for element in self.elements:
            if element.offset + element.size > len(buf):
                logger.trace("cannot unpack {} for {}.{} buffer too small {}",
                             element.name, element.block_name, element.block_version, len(buf))
                contents[element.name] = None
                continue
            s, = struct.unpack_from(element.structure, buf, element.offset)
            if element.decode:
                s = element.decode(s)
            contents[element.name] = s
        return contents

    def __iter__(self):
        for item in self.elements:
            yield item



from typing import TYPE_CHECKING
from datetime import datetime

from bxcommon.rpc import rpc_constants
from bxcommon.rpc.bx_json_rpc_request import BxJsonRpcRequest
from bxcommon.rpc.json_rpc_response import JsonRpcResponse
from bxcommon.rpc.requests.abstract_rpc_request import AbstractRpcRequest
from bxcommon.utils import config
from bxutils import logging

logger = logging.get_logger(__name__)


if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode

DEFAULT_TX_SERVICE_FILE_NAME = "tx_service_dump.csv"


class GatewayTransactionServiceRpcRequest(AbstractRpcRequest["AbstractGatewayNode"]):

    help = {
        "params": f"Optional - {rpc_constants.TX_SERVICE_FILE_NAME_PARAMS_KEY}: ",
        "description": "dump transaction service to file"
    }

    def __init__(
        self,
        request: BxJsonRpcRequest,
        node: "AbstractGatewayNode",
    ):
        self._file_name = DEFAULT_TX_SERVICE_FILE_NAME
        super().__init__(request, node)

    def validate_params(self) -> None:
        super().validate_params()
        params = self.params
        assert isinstance(params, dict)
        if rpc_constants.TX_SERVICE_FILE_NAME_PARAMS_KEY in params:
            self._file_name = params[rpc_constants.TX_SERVICE_FILE_NAME_PARAMS_KEY]

    async def process_request(self) -> JsonRpcResponse:
        path = config.get_data_file(self._file_name)
        with open(path, "w", encoding="utf-8") as f:
            f.write("transaction_hash,has_contents,has_short_id,contents_length,short_id,time_inserted\n")

            tx_service = self.node.get_tx_service()
            for tx_hash in tx_service.iter_transaction_hashes():
                transaction_cache_key = tx_service._tx_hash_to_cache_key(tx_hash)
                has_contents = transaction_cache_key in tx_service._tx_cache_key_to_contents
                has_short_id = transaction_cache_key in tx_service._tx_cache_key_to_short_ids

                contents_length = 0
                if has_contents:
                    tx_contents = tx_service._tx_cache_key_to_contents[transaction_cache_key]
                    contents_length = len(tx_contents)

                if has_short_id:
                    short_ids = tx_service._tx_cache_key_to_short_ids[transaction_cache_key]
                    for short_id in short_ids:
                        time_inserted = tx_service.get_short_id_assign_time(short_id)
                        if time_inserted == 0.0:
                            time_inserted = None
                        else:
                            time_inserted = datetime.fromtimestamp(time_inserted)

                        f.write(
                            f"{tx_hash},"
                            f"{has_contents},"
                            f"{has_short_id},"
                            f"{contents_length},"
                            f"{short_id},"
                            f"{time_inserted}\n"
                        )
                else:
                    f.write(
                        f"{tx_hash},"
                        f"{has_contents},"
                        f"{has_short_id},"
                        f"{contents_length},"
                        f"{None},"
                        f"{None}\n"
                    )
            f.close()

        return self.ok({
            "file": path
        })



from typing import List, Optional, Union

from cryptography.x509 import Certificate, ExtensionNotFound
from cryptography.x509.extensions import KeyUsage, UnrecognizedExtension, BasicConstraints

from bxcommon.models.node_type import NodeType
from bxutils import constants
from bxutils.ssl.extensions.account_id_extension import AccountIdExtension
from bxutils.ssl.extensions.extensions_object_ids import ExtensionsObjectIds
from bxutils.ssl.extensions.node_id_extension import NodeIdExtension
from bxutils.ssl.extensions.node_type_extension import NodeTypeExtension
from bxutils.ssl.extensions.node_privileges_extension import NodePrivilegesExtension

DEFAULT_KEY_USAGE = KeyUsage(
    digital_signature=True,
    key_encipherment=True,
    content_commitment=True,
    data_encipherment=False,
    key_agreement=False,
    encipher_only=False,
    decipher_only=False,
    key_cert_sign=False,
    crl_sign=False
)


def create_node_type_extension(node_type: NodeType) -> NodeTypeExtension:
    return NodeTypeExtension(node_type.name)


def create_node_id_extension(node_id: str) -> NodeIdExtension:
    return NodeIdExtension(node_id)


def create_account_id_extension(account_id: str) -> AccountIdExtension:
    return AccountIdExtension(account_id)


def create_node_privileges_extension(node_privileges: str) -> NodePrivilegesExtension:
    return NodePrivilegesExtension(node_privileges)


def get_node_type(cert: Certificate) -> Optional[NodeType]:
    try:
        # pyre-fixme[16]: `ExtensionType` has no attribute `value`.
        node_type_str = cert.extensions.get_extension_for_oid(ExtensionsObjectIds.NODE_TYPE).value.value.decode("utf-8")
        return NodeType[node_type_str]
    except ExtensionNotFound:
        return None


def get_node_id(cert: Certificate) -> Optional[str]:
    try:
        # pyre-fixme[16]: `ExtensionType` has no attribute `value`.
        return cert.extensions.get_extension_for_oid(ExtensionsObjectIds.NODE_ID).value.value.decode("utf-8")
    except ExtensionNotFound:
        return None


def get_account_id(cert: Certificate) -> Optional[str]:
    try:
        # pyre-fixme[16]: `ExtensionType` has no attribute `value`.
        return cert.extensions.get_extension_for_oid(ExtensionsObjectIds.ACCOUNT_ID).value.value.decode("utf-8")
    except ExtensionNotFound:
        return None


def get_node_privileges(cert: Certificate) -> str:
    try:
        # pyre-fixme[16]: `ExtensionType` has no attribute `value`
        return cert.extensions.get_extension_for_oid(
            ExtensionsObjectIds.NODE_PRIVILEGES
        ).value.value.decode("utf-8")
    except ExtensionNotFound:
        return constants.DEFAULT_NODE_PRIVILEGES


def get_custom_extensions(
    node_type: Optional[NodeType] = None,
    node_id: Optional[str] = None,
    account_id: Optional[str] = None,
    node_privileges: Optional[str] = None,
    ca_cert: bool = False,
    key_usage: KeyUsage = DEFAULT_KEY_USAGE
) -> List[Union[KeyUsage, UnrecognizedExtension, BasicConstraints]]:
    extensions: List[Union[KeyUsage, UnrecognizedExtension, BasicConstraints]] = [
        BasicConstraints(ca=ca_cert, path_length=None),
        key_usage
    ]
    if node_type is not None:
        extensions.append(create_node_type_extension(node_type))
    if node_id is not None:
        extensions.append(create_node_id_extension(node_id))
    if account_id is not None:
        extensions.append(create_account_id_extension(account_id))
    if node_privileges is not None:
        extensions.append(create_node_privileges_extension(node_privileges))
    return extensions



import struct

from bxcommon.utils.blockchain_utils.btc.btc_common_utils import btc_varint_to_int
from bxcommon.utils.blockchain_utils.btc.btc_object_hash import BtcObjectHash
from bxgateway.btc_constants import BTC_HDR_COMMON_OFF, BTC_SHA_HASH_LEN
from bxgateway.messages.btc.btc_message import BtcMessage
from bxgateway.messages.btc.btc_message_type import BtcMessageType
from bxgateway.messages.btc.btc_messages_util import pack_int_to_btc_varint


class DataBtcMessage(BtcMessage):
    def __init__(self, magic=None, version=None, hashes=None, hash_stop=None, command=None, buf=None):
        if hashes is None:
            hashes = []
        if buf is None:
            buf = bytearray(BTC_HDR_COMMON_OFF + 9 + (len(hashes) + 1) * 32)
            self.buf = buf

            off = BTC_HDR_COMMON_OFF
            struct.pack_into("<I", buf, off, version)
            off += 4
            off += pack_int_to_btc_varint(len(hashes), buf, off)

            for hash_val in hashes:
                buf[off:off + 32] = hash_val.get_big_endian()
                off += 32

            buf[off:off + 32] = hash_stop.get_big_endian()
            off += 32

            BtcMessage.__init__(self, magic, command, off - BTC_HDR_COMMON_OFF, buf)
        else:
            self.buf = buf
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._version = self._hash_count = self._hashes = self._hash_stop = None

    def version(self):
        if self._version is None:
            self._version, = struct.unpack_from("<I", self.buf, BTC_HDR_COMMON_OFF)
        return self._version

    def hash_count(self):
        if self._hash_count is None:
            off = BTC_HDR_COMMON_OFF + 4
            self._hash_count, size = btc_varint_to_int(self.buf, off)

        return self._hash_count

    def __iter__(self):
        off = BTC_HDR_COMMON_OFF + 4  # For the version field.
        b_count, size = btc_varint_to_int(self.buf, off)
        off += size

        for i in range(b_count):
            yield BtcObjectHash(buf=self.buf, offset=off, length=BTC_SHA_HASH_LEN)
            off += 32

    def hash_stop(self):
        return BtcObjectHash(buf=self.buf, offset=BTC_HDR_COMMON_OFF + self.payload_len() - 32, length=BTC_SHA_HASH_LEN)


class GetHeadersBtcMessage(DataBtcMessage):
    MESSAGE_TYPE = BtcMessageType.GET_HEADERS

    def __init__(self, magic=None, version=None, hashes=None, hash_stop=None, buf=None):
        if hashes is None:
            hashes = []
        super(GetHeadersBtcMessage, self).__init__(magic, version, hashes, hash_stop, self.MESSAGE_TYPE, buf)


class GetBlocksBtcMessage(DataBtcMessage):
    MESSAGE_TYPE = BtcMessageType.GET_BLOCKS

    def __init__(self, magic=None, version=None, hashes=None, hash_stop=None, buf=None):
        if hashes is None:
            hashes = []
        super(GetBlocksBtcMessage, self).__init__(magic, version, hashes, hash_stop, self.MESSAGE_TYPE, buf)



import blxr_rlp as rlp
from bxgateway.messages.eth.protocol.eth_protocol_message import EthProtocolMessage
from bxgateway.messages.eth.protocol.eth_protocol_message_type import EthProtocolMessageType


class NodeDataEthProtocolMessage(EthProtocolMessage):
    msg_type = EthProtocolMessageType.NODE_DATA

    fields = [("raw_data", rlp.sedes.raw)]

    def get_raw_data(self):
        return self.get_field_value("raw_data")


from typing import TYPE_CHECKING, cast

from bxcommon.feed.feed_manager import FeedManager
from bxcommon.rpc.abstract_ws_rpc_handler import AbstractWsRpcHandler
from bxcommon.rpc.requests.transaction_status_rpc_request import TransactionStatusRpcRequest
from bxcommon.rpc.rpc_request_type import RpcRequestType
from bxgateway.rpc.requests.add_blockchain_peer_rpc_request import AddBlockchainPeerRpcRequest
from bxgateway.rpc.requests.bdn_performance_rpc_request import BdnPerformanceRpcRequest
from bxgateway.rpc.requests.gateway_blxr_transaction_rpc_request import GatewayBlxrTransactionRpcRequest
from bxgateway.rpc.requests.gateway_memory_usage_report_rpc_request import GatewayMemoryUsageRpcRequest
from bxgateway.rpc.requests.gateway_status_rpc_request import GatewayStatusRpcRequest
from bxgateway.rpc.requests.gateway_stop_rpc_request import GatewayStopRpcRequest
from bxgateway.rpc.requests.gateway_memory_rpc_request import GatewayMemoryRpcRequest
from bxgateway.rpc.requests.gateway_peers_rpc_request import GatewayPeersRpcRequest
from bxgateway.rpc.requests.gateway_transaction_service_rpc_request import GatewayTransactionServiceRpcRequest
from bxgateway.rpc.requests.quota_usage_rpc_request import QuotaUsageRpcRequest
from bxgateway.rpc.requests.gateway_blxr_call_rpc_request import GatewayBlxrCallRpcRequest
from bxgateway.rpc.requests.remove_blockchain_peer_rpc_request import RemoveBlockchainPeerRpcRequest
from bxcommon.rpc.requests.subscribe_rpc_request import SubscribeRpcRequest
from bxcommon.rpc.requests.unsubscribe_rpc_request import UnsubscribeRpcRequest

from bxutils import logging
from bxutils.encoding.json_encoder import Case

if TYPE_CHECKING:
    # noinspection PyUnresolvedReferences
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxgateway.connections.abstract_gateway_node import AbstractGatewayNode
    from bxcommon.connections.abstract_node import AbstractNode

logger = logging.get_logger(__name__)


class GatewayWsHandler(AbstractWsRpcHandler):

    def __init__(self, node: "AbstractGatewayNode", feed_manager: FeedManager, case: Case) -> None:
        super().__init__(node, feed_manager, case)
        self.request_handlers = {
            RpcRequestType.BLXR_TX: GatewayBlxrTransactionRpcRequest,
            RpcRequestType.BLXR_ETH_CALL: GatewayBlxrCallRpcRequest,
            RpcRequestType.GATEWAY_STATUS: GatewayStatusRpcRequest,
            RpcRequestType.STOP: GatewayStopRpcRequest,
            RpcRequestType.MEMORY: GatewayMemoryRpcRequest,
            RpcRequestType.PEERS: GatewayPeersRpcRequest,
            RpcRequestType.BDN_PERFORMANCE: BdnPerformanceRpcRequest,
            RpcRequestType.QUOTA_USAGE: QuotaUsageRpcRequest,
            RpcRequestType.MEMORY_USAGE: GatewayMemoryUsageRpcRequest,
            RpcRequestType.TX_STATUS: TransactionStatusRpcRequest,
            RpcRequestType.TX_SERVICE: GatewayTransactionServiceRpcRequest,
            RpcRequestType.ADD_BLOCKCHAIN_PEER: AddBlockchainPeerRpcRequest,
            RpcRequestType.REMOVE_BLOCKCHAIN_PEER: RemoveBlockchainPeerRpcRequest,
            RpcRequestType.SUBSCRIBE: SubscribeRpcRequest,
            RpcRequestType.UNSUBSCRIBE: UnsubscribeRpcRequest,
        }



import functools


def deprecated(func):
    """This is a decorator which can be used to mark functions
    as deprecated."""
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        return func(*args, **kwargs)
    return new_func



from bxcommon.utils.stats.statistics_event_service import StatisticsEventService
from bxutils import logging
from bxutils.logging.log_record_type import LogRecordType


class _ConnectionStatisticsService(StatisticsEventService):
    def __init__(self) -> None:
        super(_ConnectionStatisticsService, self).__init__()
        self.name = "ConnectionState"
        self.logger = logging.get_logger(LogRecordType.ConnectionState, __name__)


connection_stats = _ConnectionStatisticsService()



import time
from collections import deque
from typing import Set, Optional

from bxcommon import constants
from bxcommon.utils import memory_utils
from bxcommon.utils.memory_utils import SpecialMemoryProperties, SpecialTuple
from bxutils import logging

logger = logging.get_logger(__name__)


class OutputBuffer(SpecialMemoryProperties):
    """
    There are three key functions on the outputbuffer read interface. This should also
    be implemented by the cut through sink interface.
      - has_more_bytes(): Whether or not there are more bytes in this buffer.
      - get_buffer(): some bytes to send in the outputbuffer
      - advance_buffer(): Advances the buffer by some number of bytes
    """
    EMPTY = memoryview(bytearray(0))  # The empty outputbuffer

    def __init__(self, min_size=None, max_hold_time=None, enable_buffering=False) -> None:
        if min_size is None:
            min_size = constants.OUTPUT_BUFFER_MIN_SIZE
        if max_hold_time is None:
            max_hold_time = constants.OUTPUT_BUFFER_BATCH_MAX_HOLD_TIME

        self.enable_buffering = enable_buffering

        # A deque of memoryview objects representing the raw memoryviews of the messages
        # that are being sent on the outputbuffer.
        self.output_msgs = deque()

        # Offset into the first message of the output_msgs
        self.index = 0

        # The total sum of all of the messages in the outputbuffer
        self.length = 0

        self.min_size = min_size
        # how long we hold onto messages for batching in seconds
        self.max_hold_time = max_hold_time
        self.last_memview = None
        self.last_bytearray: Optional[bytearray] = None
        # size of the last valid memoryview
        self.valid_len = 0
        self.last_bytearray_create_time = None

    def __len__(self):
        return self.length

    def get_buffer(self):
        """
        Gets a non-empty memoryview buffer
        :return: top output message on buffer
        """
        now = time.time()

        if self.enable_buffering and \
                self.last_bytearray is not None and \
                now - self.last_bytearray_create_time >= self.max_hold_time:
            self.flush()

        if not self.output_msgs:
            return OutputBuffer.EMPTY

        return memoryview(self.output_msgs[0])[self.index:]

    def advance_buffer(self, num_bytes: int):
        if num_bytes < 0:
            raise ValueError("Num_bytes must be a positive integer.")

        if (not self.output_msgs and num_bytes > 0) or (self.index + num_bytes) > len(self.output_msgs[0]):
            if self.output_msgs:
                output_message = self.output_msgs[0].tobytes().hex()
            else:
                output_message = None
            raise ValueError(
                "Index cannot be larger than length of first message. Message: {}, Index: {}, Bytes: {}".format(
                    output_message, self.index, num_bytes)
            )

        self.index += num_bytes
        self.length -= num_bytes

        if self.index == len(self.output_msgs[0]):
            self.output_msgs.popleft()
            self.index = 0

    def at_msg_boundary(self):
        return self.index == 0

    def enqueue_msgbytes(self, msg_bytes):
        if not isinstance(msg_bytes, (bytearray, memoryview)):
            raise ValueError("Msg_bytes must be a bytearray. The type given was a {}".format(type(msg_bytes)))

        length = len(msg_bytes)

        if not self.enable_buffering:
            self.output_msgs.append(msg_bytes)
        elif length + self.valid_len > self.min_size:
            if self.last_bytearray is not None:
                self.flush()
            self.output_msgs.append(msg_bytes)
        else:
            now = time.time()
            if self.last_bytearray is None:
                self.last_bytearray = bytearray(self.min_size)
                self.last_memview = memoryview(self.last_bytearray)
                self.last_bytearray[:length] = msg_bytes
                self.valid_len = length
                self.last_bytearray_create_time = now
            else:
                if self.last_bytearray_create_time is None:
                    raise ValueError("last_bytearray_create_time cannot be None")
                self.last_bytearray[self.valid_len:self.valid_len + length] = msg_bytes
                self.valid_len += length

                if now - self.last_bytearray_create_time > self.max_hold_time:
                    self.flush()

        self.length += len(msg_bytes)

    def prepend_msgbytes(self, msg_bytes):
        if not isinstance(msg_bytes, (bytearray, memoryview)):
            raise ValueError("Msg_bytes must be a bytearray.")

        if self.index == 0:
            self.output_msgs.appendleft(msg_bytes)
        else:
            prev_msg = self.output_msgs.popleft()
            self.output_msgs.appendleft(msg_bytes)
            self.output_msgs.appendleft(prev_msg)

        self.length += len(msg_bytes)

    def has_more_bytes(self):
        return self.length != 0

    # TODO: @soumya this is called every 200ms. This needs some future cleanup; possibly in the alarm data structure.
    # Consult Nagle algorithm before implementing improvements.
    def flush(self):
        if self.last_bytearray is None:
            return

        self.output_msgs.append(self.last_memview[:self.valid_len])
        self.last_bytearray_create_time = None
        self.last_bytearray = None
        self.last_memview = None
        self.valid_len = 0

    def safe_empty(self):
        """
        Removes all bytes in OutputBuffer that are not in the current message boundary.
        """
        self.flush()
        if self.output_msgs and self.index:
            first = self.output_msgs.popleft()
            self.output_msgs.clear()
            self.output_msgs.append(first)
            self.length = len(first) - self.index
        else:
            self.output_msgs.clear()
            self.length = 0

    def special_memory_size(self, ids: Optional[Set[int]] = None) -> SpecialTuple:
        return memory_utils.get_special_size(self.output_msgs, ids=ids)



from typing import List, NamedTuple, Optional, Union

from bxcommon.utils.object_hash import Sha256Hash


class TransactionInfo(NamedTuple):
    hash: Optional[Sha256Hash]
    contents: Optional[Union[bytearray, memoryview]]
    short_id: Optional[int]


class TransactionSearchResult(NamedTuple):
    found: List[TransactionInfo]
    missing: List[TransactionInfo]



# pylint: disable=too-many-lines

import functools
import time
import typing
from collections import defaultdict, OrderedDict, Counter
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from functools import reduce
from typing import List, Tuple, Generator, Optional, Union, Dict, Set, Any, Iterator, TYPE_CHECKING

from prometheus_client import Gauge

from bxcommon import constants
from bxcommon.messages.bloxroute import short_ids_serializer
from bxcommon.messages.bloxroute.tx_service_sync_txs_message import TxServiceSyncTxsMessage
from bxcommon.models.transaction_flag import TransactionFlag
from bxcommon.models.transaction_info import TransactionSearchResult, TransactionInfo
from bxcommon.models.transaction_key import TransactionKey, TransactionCacheKeyType
from bxcommon.utils import memory_utils, convert
from bxcommon.utils.crypto import SHA256_HASH_LEN
from bxcommon.utils.deprecated import deprecated
from bxcommon.utils.expiration_queue import ExpirationQueue
from bxcommon.utils.memory_utils import ObjectSize, SizeType
from bxcommon.utils.object_hash import Sha256Hash
from bxcommon.utils.stats import hooks
from bxcommon.utils.stats.transaction_stat_event_type import TransactionStatEventType
from bxcommon.utils.stats.transaction_statistics_service import tx_stats
from bxutils import log_messages
from bxutils import logging, utils
from bxutils.encoding import json_encoder
from bxutils.logging.log_record_type import LogRecordType

if TYPE_CHECKING:
    # pylint: disable=ungrouped-imports,cyclic-import
    from bxcommon.connections.abstract_node import AbstractNode
else:
    tpe_Sha256 = Any

logger = logging.get_logger(__name__)
logger_memory_cleanup = logging.get_logger(LogRecordType.BlockCleanup, __name__)
logger_tx_histogram = logging.get_logger(LogRecordType.TransactionHistogram, __name__)
performance_logger = logging.get_logger(LogRecordType.PerformanceTroubleshooting, __name__)

total_cached_transactions = Gauge(
    "cached_transaction_total",
    "Number of cached transactions",
    ("network",)
)
total_cached_transactions_size = Gauge(
    "cached_transaction_bytes",
    "Size of cached transactions",
    ("network",)
)


def wrap_sha256(transaction_hash: Union[bytes, bytearray, memoryview, Sha256Hash]) -> Sha256Hash:
    if isinstance(transaction_hash, Sha256Hash):
        return transaction_hash

    # pyre-fixme[25]: Assertion will always fail.
    if isinstance(transaction_hash, (bytes, bytearray, memoryview)):
        return Sha256Hash(binary=transaction_hash)

    return Sha256Hash(binary=convert.hex_to_bytes(transaction_hash))


class TxRemovalReason(Enum):
    UNKNOWN = "Unknown"
    BLOCK_CLEANUP = "BlockCleanup"
    EXTENSION_BLOCK_CLEANUP = "ExtensionBlockCleanup"
    MEMORY_LIMIT = "MemoryLimit"
    EXPIRATION = "Expiration"
    MEMPOOL_SYNC = "MempoolSync"


@dataclass
class TransactionServiceStats:
    short_id_count: int = 0
    transaction_content_count: int = 0
    transactions_removed_by_memory_limit: int = 0
    transaction_contents_size: int = 0

    def __sub__(self, other) -> "TransactionServiceStats":
        if not isinstance(other, TransactionServiceStats):
            raise TypeError(f"Cannot subtract object of type {type(other)} from TransactionServiceStats.")

        return TransactionServiceStats(
            self.short_id_count - other.short_id_count,
            self.transaction_content_count - other.transaction_content_count,
            self.transactions_removed_by_memory_limit - other.transactions_removed_by_memory_limit,
            self.transaction_contents_size - other.transaction_contents_size
        )


class TransactionFromBdnGatewayProcessingResult(typing.NamedTuple):
    ignore_seen: bool = False
    existing_short_id: bool = False
    assigned_short_id: bool = False
    existing_contents: bool = False
    set_content: bool = False


class TxSyncMsgProcessingItem(typing.NamedTuple):
    hash: Optional[Sha256Hash] = None
    content_length: int = 0
    short_ids: List[int] = []
    transaction_flag_types: List[TransactionFlag] = []


# pylint: disable=too-many-public-methods
class TransactionService:
    """
    Service for managing transaction mappings.
    In this class, we assume that no more than MAX_ID unassigned transactions exist at a time.

    Constants
    ---------
    MAX_ID: maximum short id value (e.g. number of bits in a short id)
    SHORT_ID_SIZE: number of bytes in a short id, must match TxMessage


    Attributes
    ----------
    node: reference to node holding transaction service reference
    tx_assign_alarm_scheduled: if an alarm to expire a batch of short ids is currently active
    network_num: network number that current transaction service serves
    _tx_cache_key_to_short_ids: mapping of transaction long hashes to (potentially multiple) short ids
    _short_id_to_tx_cache_key: mapping of short id to transaction long hashes
    _short_id_to_tx_flag: mapping of short id to transaction flag type
    _tx_cache_key_to_contents: mapping of transaction long hashes to transaction contents
    _tx_assignment_expire_queue: expiration time of short ids
    """

    node: "AbstractNode"
    tx_assign_alarm_scheduled: bool
    tx_content_without_sid_alarm_scheduled: bool
    network_num: int
    _short_id_to_tx_cache_key: Dict[int, TransactionCacheKeyType]
    _short_id_to_tx_flag: Dict[int, TransactionFlag]
    _tx_cache_key_to_contents: Dict[TransactionCacheKeyType, Union[bytearray, memoryview]]
    _tx_cache_key_to_short_ids: Dict[TransactionCacheKeyType, Set[int]]
    _tx_assignment_expire_queue: ExpirationQueue[int]
    _tx_hash_to_time_removed: OrderedDict
    _short_id_to_time_removed: OrderedDict
    tx_hashes_without_short_id: ExpirationQueue[Sha256Hash]
    tx_hashes_without_content: ExpirationQueue[Sha256Hash]  # but has short ID

    MAX_ID = 2 ** 32
    SHORT_ID_SIZE = 4
    DEFAULT_FINAL_TX_CONFIRMATIONS_COUNT = 24

    ESTIMATED_TX_HASH_AND_SHORT_ID_ITEM_SIZE = 376
    ESTIMATED_TX_HASH_ITEM_SIZE = 312
    ESTIMATED_SHORT_ID_EXPIRATION_ITEM_SIZE = 88
    ESTIMATED_TX_HASH_NOT_SEEN_IN_BLOCK_ITEM_SIZE = 312

    def __init__(self, node: "AbstractNode", network_num: int) -> None:
        """
        Constructor
        :param node: reference to node object
        :param network_num: network number
        """
        if node is None:
            raise ValueError("Node is required")

        if network_num is None:
            raise ValueError("Network number is required")

        self.node = node
        self.network_num = network_num

        self.tx_assign_alarm_scheduled = False
        self.tx_content_without_sid_alarm_scheduled = False
        self.tx_without_content_alarm_scheduled = False

        self._tx_cache_key_to_short_ids = defaultdict(set)
        self._short_id_to_tx_flag = {}
        self._short_id_to_tx_cache_key = {}
        self._tx_cache_key_to_contents = {}
        self._tx_assignment_expire_queue = ExpirationQueue(node.opts.sid_expire_time)
        self.tx_hashes_without_short_id = ExpirationQueue(constants.TX_CONTENT_NO_SID_EXPIRE_S)
        self.tx_hashes_without_content = ExpirationQueue(constants.TX_CONTENT_NO_SID_EXPIRE_S)
        self.network = None
        if self.network_num in self.node.opts.blockchain_networks:
            self.network = self.node.opts.blockchain_networks[self.network_num]

        self._tx_hash_to_time_removed = OrderedDict()
        self._short_id_to_time_removed = OrderedDict()

        self._final_tx_confirmations_count = self._get_final_tx_confirmations_count()
        self._tx_content_memory_limit = self._get_tx_contents_memory_limit()
        logger.debug("Memory limit for transaction service by network number {} is {} bytes.",
                     self.network_num, self._tx_content_memory_limit)
        self._removed_txs_expiration_time_s = self._get_removed_transactions_history_expiration_time_s()

        # short ids seen in block ordered by them block hash
        self._short_ids_seen_in_block: Dict[Sha256Hash, List[int]] = OrderedDict()
        self._total_tx_contents_size = 0
        self._total_tx_removed_by_memory_limit = 0

        self._last_transaction_stats = TransactionServiceStats()
        self._removed_short_ids = set()
        if node.opts.dump_removed_short_ids:
            self.node.alarm_queue.register_alarm(
                constants.DUMP_REMOVED_SHORT_IDS_INTERVAL_S,
                self._dump_removed_short_ids
            )
        self.node.alarm_queue.register_alarm(
            constants.REMOVED_TRANSACTIONS_HISTORY_CLEANUP_INTERVAL_S,
            self._cleanup_removed_transactions_history
        )

        self.total_cached_transactions = total_cached_transactions.labels(network_num)
        self.total_cached_transactions.set_function(
            functools.partial(len, self._tx_cache_key_to_contents)
        )
        self.total_cached_transactions_size = total_cached_transactions_size.labels(network_num)
        self.total_cached_transactions_size.set_function(
            functools.partial(utils.identity, self._total_tx_contents_size)
        )

    def get_short_id_transaction_type(self, short_id: int) -> TransactionFlag:
        if short_id in self._short_id_to_tx_flag:
            return self._short_id_to_tx_flag[short_id]
        else:
            return TransactionFlag.NO_FLAGS

    def set_short_id_transaction_type(self, short_id: int, tx_flag: TransactionFlag) -> None:
        if TransactionFlag.PAID_TX & tx_flag:
            self._short_id_to_tx_flag[short_id] = tx_flag

    def get_short_id_by_key(self, transaction_key: TransactionKey) -> int:
        """
        Fetches a single short id for transaction. If the transaction has multiple short id mappings, just gets
        the first one.
        :param transaction_key: transaction key
        :return: short id
        """
        return next(iter(self.get_short_ids_by_key(transaction_key)))

    def get_short_ids_by_key(self, transaction_key: TransactionKey) -> Set[int]:
        """
        Fetches all short ids for a given transactions
        :param transaction_key: transaction key
        :return: set of short ids
        """

        if transaction_key.transaction_cache_key in self._tx_cache_key_to_short_ids:
            return self._tx_cache_key_to_short_ids[transaction_key.transaction_cache_key]
        else:
            return constants.NULL_TX_SIDS

    def get_transaction(self, short_id: int) -> TransactionInfo:
        """
        Fetches transaction info for a given short id.
        Results might be None.
        :param short_id:
        :return: transaction hash, transaction contents.
        """
        if short_id in self._short_id_to_tx_cache_key:
            transaction_cache_key = self._short_id_to_tx_cache_key[short_id]
            if transaction_cache_key in self._tx_cache_key_to_contents:
                transaction_contents = self._tx_cache_key_to_contents[transaction_cache_key]
                return TransactionInfo(self._tx_cache_key_to_hash(transaction_cache_key),
                                       transaction_contents,
                                       short_id)
            else:
                return TransactionInfo(self._tx_cache_key_to_hash(transaction_cache_key), None, short_id)
        else:
            return TransactionInfo(None, None, short_id)

    def get_missing_transactions(
        self, short_ids: List[int]
    ) -> Tuple[bool, List[int], List[Sha256Hash]]:
        unknown_tx_sids = []
        unknown_tx_hashes = []
        has_missing = False
        for short_id in short_ids:
            transaction_cache_key = self._short_id_to_tx_cache_key.get(short_id, None)
            if transaction_cache_key is None:
                unknown_tx_sids.append(short_id)
                has_missing = True
                continue
            transaction_key = self.get_transaction_key(None, transaction_cache_key)
            if not self.has_transaction_contents_by_key(transaction_key):
                unknown_tx_hashes.append(transaction_key.transaction_hash)
                has_missing = True
        return has_missing, unknown_tx_sids, unknown_tx_hashes

    def get_transaction_by_key(self, transaction_key: TransactionKey) -> Optional[Union[bytearray, memoryview]]:
        """
        Fetches transaction contents for a given transaction key.
        Results might be None.
        :param transaction_key: transaction key
        :return: transaction contents.
        """
        if transaction_key.transaction_cache_key in self._tx_cache_key_to_contents:
            return self._tx_cache_key_to_contents[transaction_key.transaction_cache_key]

        return None

    def get_transactions(
        self,
        serialized_short_ids: Optional[bytearray] = None
    ) -> TransactionSearchResult:
        """
        Fetches all transaction info for a set of short ids.
        Short ids without a transaction entry will be omitted.
        Function allows to pass a single short id or serialized list of short ids
        :param serialized_short_ids: instance of get transactions message
        :return: list of found and missing short ids
        """

        assert serialized_short_ids is not None
        short_ids = short_ids_serializer.deserialize_short_ids(serialized_short_ids)

        found = []
        missing = []
        for short_id in short_ids:
            if short_id in self._short_id_to_tx_cache_key:
                transaction_cache_key = self._short_id_to_tx_cache_key[short_id]
                if transaction_cache_key in self._tx_cache_key_to_contents:
                    found.append(TransactionInfo(
                        self._tx_cache_key_to_hash(transaction_cache_key),
                        self._tx_cache_key_to_contents[transaction_cache_key],
                        short_id
                    ))
                else:
                    missing.append(TransactionInfo(
                        self._tx_cache_key_to_hash(transaction_cache_key),
                        None,
                        short_id
                    ))
                    logger.trace("Short id {} was requested but is unknown.", short_id)
            else:
                missing.append(TransactionInfo(None, None, short_id))
                logger.trace("Short id {} was requested but is unknown.", short_id)

        return TransactionSearchResult(found, missing)

    def get_tx_hash_to_contents_len(self) -> int:
        return len(self._tx_cache_key_to_contents)

    def get_short_id_count(self) -> int:
        return len(self._short_id_to_tx_cache_key)

    @deprecated
    def has_transaction_contents(self, transaction_hash: Sha256Hash) -> bool:
        """
        Checks if transaction contents is available in transaction service cache

        :param transaction_hash: transaction hash
        :return: Boolean indicating if transaction contents exists
        """
        return self.has_transaction_contents_by_key(self.get_transaction_key(transaction_hash))

    @deprecated
    def has_transaction_contents_by_cache_key(self, cache_key: TransactionCacheKeyType) -> bool:
        """
        Checks if transaction contents is available in transaction service cache

        :param cache_key: transaction cache key
        :return: Boolean indicating if transaction contents exists
        """
        return cache_key in self._tx_cache_key_to_contents

    def has_transaction_contents_by_key(self, transaction_key: TransactionKey) -> bool:
        """
        Checks if transaction contents is available in transaction service cache

        :param transaction_key: transaction key
        :return: Boolean indicating if transaction contents exists
        """
        return transaction_key.transaction_cache_key in self._tx_cache_key_to_contents

    @deprecated
    def has_transaction_short_id(self, transaction_hash: Sha256Hash) -> bool:
        """
        Checks if transaction short id is available in transaction service cache

        :param transaction_hash: transaction hash
        :return: Boolean indicating if transaction short id exists
        """
        return self.has_transaction_short_id_by_key(self.get_transaction_key(transaction_hash))

    def has_transaction_short_id_by_key(self, transaction_key: TransactionKey) -> bool:
        """
        Checks if transaction short id is available in transaction service cache

        :param transaction_key: transaction key
        :return: Boolean indicating if transaction short id exists
        """
        return transaction_key.transaction_cache_key in self._tx_cache_key_to_short_ids

    def has_short_id(self, short_id: int) -> bool:
        """
        Checks if short id is stored in transaction service cache
        :param short_id: transaction short id
        :return: Boolean indicating if short id is found in cache
        """
        return short_id in self._short_id_to_tx_cache_key

    def has_cache_key_no_sid_entry(self, transaction_hash: Sha256Hash) -> bool:
        return transaction_hash in self.tx_hashes_without_short_id.queue

    @deprecated
    def removed_transaction(self, transaction_hash: Sha256Hash) -> bool:
        """
        Check if transaction was seen and removed from cache
        :param transaction_hash: transaction hash
        :return: Boolean indicating in transaction was seen
        """
        return self.removed_transaction_by_key(self.get_transaction_key(transaction_hash))

    @deprecated
    def removed_transaction_by_cache_key(self, cache_key: TransactionCacheKeyType) -> bool:
        """
        Check if transaction was seen and removed from cache
        :param cache_key: transaction cache key
        :return: Boolean indicating in transaction was seen
        """
        return cache_key in self._tx_hash_to_time_removed

    def removed_transaction_by_key(self, transaction_key: TransactionKey) -> bool:
        """
        Check if transaction was seen and removed from cache
        :param transaction_key: transaction key
        :return: Boolean indicating in transaction was seen
        """
        return transaction_key.transaction_cache_key in self._tx_hash_to_time_removed

    @deprecated
    def remove_sid_by_tx_hash(self, transaction_hash: Sha256Hash) -> None:
        self.remove_sid_by_key(self.get_transaction_key(transaction_hash))

    def remove_sid_by_key(self, transaction_key: TransactionKey) -> None:
        if transaction_key.transaction_cache_key in self._tx_cache_key_to_contents:
            logger.trace("attempted to clear sid, for transaction {} with content, ignore", transaction_key)
        else:
            if transaction_key.transaction_cache_key in self._tx_cache_key_to_short_ids:
                short_ids = self._tx_cache_key_to_short_ids[transaction_key.transaction_cache_key]
                del self._tx_cache_key_to_short_ids[transaction_key.transaction_cache_key]
                for short_id in short_ids:
                    del self._short_id_to_tx_cache_key[short_id]
                    self._tx_assignment_expire_queue.remove(short_id)

    @deprecated
    def assign_short_id(
        self,
        transaction_hash: Sha256Hash,
        short_id: int,
        transaction_cache_key: Optional[TransactionCacheKeyType] = None
    ) -> None:
        return self.assign_short_id_by_key(
            self.get_transaction_key(transaction_hash, transaction_cache_key),
            short_id
        )

    def assign_short_id_by_key(
        self,
        transaction_key: TransactionKey,
        short_id: int,
    ) -> None:
        """
        Adds short id mapping for transaction and schedules an alarm to cleanup entry on expiration.
        :param transaction_key: transaction key
        :param short_id: short id to be mapped to transaction
        """
        has_contents = transaction_key.transaction_cache_key in self._tx_cache_key_to_contents
        self.assign_short_id_base_by_key(transaction_key, short_id, has_contents, True)

    @deprecated
    def assign_short_id_base(
        self,
        transaction_hash: Sha256Hash,
        transaction_cache_key: Optional[TransactionCacheKeyType],
        short_id: int,
        has_contents: bool,
        call_to_assign_short_id: bool
    ) -> None:
        return self.assign_short_id_base_by_key(
            self.get_transaction_key(transaction_hash, transaction_cache_key),
            short_id,
            has_contents,
            call_to_assign_short_id
        )

    def assign_short_id_base_by_key(
        self,
        transaction_key: TransactionKey,
        short_id: int,
        has_contents: bool,
        call_to_assign_short_id: bool
    ) -> None:
        """
        Base method to assign short id for a transaction

        :param transaction_key: transaction key object
        :param short_id: transaction short id
        :param has_contents: flag indicating if content already exists in cache for given transaction
        :param call_to_assign_short_id: flag indicating if method should make a call to assign short id form Python code
        :return:
        """
        if short_id == constants.NULL_TX_SID:
            # TODO: this should be an assertion; requires testing
            logger.warning(log_messages.ATTEMPTED_TO_ASSIGN_NULL_SHORT_ID_TO_TX_HASH, transaction_key)
            return
        logger.trace("Assigning sid {} to transaction {}", short_id, transaction_key)

        if not has_contents:
            self.tx_hashes_without_content.add(transaction_key.transaction_hash)
            if not self.tx_without_content_alarm_scheduled:
                self.node.alarm_queue.register_alarm(
                    constants.TX_CONTENT_NO_SID_EXPIRE_S,
                    self.expire_sid_without_content
                )
                self.tx_without_content_alarm_scheduled = True

        if call_to_assign_short_id:
            self._tx_cache_key_to_short_ids[transaction_key.transaction_cache_key].add(short_id)
            self._short_id_to_tx_cache_key[short_id] = transaction_key.transaction_cache_key
        self._tx_assignment_expire_queue.add(short_id)
        self.tx_hashes_without_short_id.remove(transaction_key.transaction_hash)

        if not self.tx_assign_alarm_scheduled:
            self.node.alarm_queue.register_alarm(self.node.opts.sid_expire_time, self.expire_old_assignments)
            self.tx_assign_alarm_scheduled = True

    def set_final_tx_confirmations_count(self, val: int) -> None:
        self._final_tx_confirmations_count = val

    @deprecated
    def set_transaction_contents(
        self, transaction_hash: Sha256Hash, transaction_contents: Union[bytearray, memoryview],
        transaction_cache_key: Optional[TransactionCacheKeyType] = None
    ):
        return self.set_transaction_contents_by_key(
            self.get_transaction_key(transaction_hash, transaction_cache_key),
            transaction_contents
        )

    def set_transaction_contents_by_key(
        self, transaction_key: TransactionKey, transaction_contents: Union[bytearray, memoryview],
    ):
        """
        Adds transaction contents to transaction service cache with lookup key by transaction hash

        :param transaction_key: transaction key object
        :param transaction_contents: transaction contents bytes
        """
        previous_size = 0

        if transaction_key.transaction_cache_key in self._tx_cache_key_to_contents:
            previous_size = len(self._tx_cache_key_to_contents[transaction_key.transaction_cache_key])
        has_short_id = transaction_key.transaction_cache_key in self._tx_cache_key_to_short_ids

        self.set_transaction_contents_base_by_key(
            transaction_key,
            has_short_id,
            previous_size,
            True,
            transaction_contents,
            None
        )

    def set_transaction_contents_base_by_key(
        self,
        transaction_key: TransactionKey,
        has_short_id: bool,
        previous_size: int,
        call_set_contents: bool,
        transaction_contents: Optional[Union[bytearray, memoryview]] = None,
        transaction_contents_length: Optional[int] = None
    ) -> None:
        """
        Adds transaction contents to transaction service cache with lookup key by transaction hash

        :param transaction_key:
        :param has_short_id: flag indicating if cache already has short id for given transaction
        :param previous_size: previous size of transaction contents if already exists
        :param call_set_contents: flag indicating if method should make a call to set content form Python code
        :param transaction_contents: transaction contents bytes
        :param transaction_contents_length: if the transaction contents bytes not available, just send the length
        """
        if not has_short_id:
            self.tx_hashes_without_short_id.add(transaction_key.transaction_hash)
            if not self.tx_content_without_sid_alarm_scheduled:
                self.node.alarm_queue.register_alarm(constants.TX_CONTENT_NO_SID_EXPIRE_S,
                                                     self.expire_content_without_sid)
                self.tx_content_without_sid_alarm_scheduled = True

        self.tx_hashes_without_content.remove(transaction_key.transaction_hash)

        if transaction_contents is not None:
            self._total_tx_contents_size += len(transaction_contents) - previous_size
            if call_set_contents:
                self._tx_cache_key_to_contents[transaction_key.transaction_cache_key] = transaction_contents
        elif transaction_contents_length is not None:
            self._total_tx_contents_size += transaction_contents_length - previous_size
        else:
            logger.debug("both transaction contents and transaction contents length are missing.")

        self._memory_limit_clean_up()

    @deprecated
    def remove_transaction_by_tx_hash(
        self,
        transaction_hash: Sha256Hash,
        force: bool = True,
        assume_no_sid: bool = False
    ) -> Optional[Set[int]]:
        return self.remove_transaction_by_key(self.get_transaction_key(transaction_hash), force, assume_no_sid)

    def remove_transaction_by_key(
        self,
        transaction_key: TransactionKey,
        force: bool = True,
        assume_no_sid: bool = False
    ) -> Optional[Set[int]]:
        """
        Clean up mapping. Removes transaction contents and mapping.
        :param transaction_key: tx key to clean up
        :param force: when false, cleanup will ignore tx / sids marked with tx flag
        :param assume_no_sid: cleanup request will be ignored if the tx has short ids
        """
        removed_sids = 0
        removed_txns = 0

        if transaction_key.transaction_cache_key in self._tx_cache_key_to_short_ids:
            time_removed = time.time()
            short_ids = self._tx_cache_key_to_short_ids[transaction_key.transaction_cache_key]
            if assume_no_sid and short_ids:
                logger.trace("removal of tx {} aborted, ", transaction_key)
                return None
            short_id_flags = [
                self._short_id_to_tx_flag.get(sid, TransactionFlag.NO_FLAGS)
                for sid in short_ids
            ]
            tx_flag = reduce(lambda x, y: x | y, short_id_flags)
            if TransactionFlag.PAID_TX in tx_flag and not force:
                return None
            else:
                short_ids = self._tx_cache_key_to_short_ids.pop(transaction_key.transaction_cache_key)

            for _, short_id in zip(short_id_flags, short_ids):
                tx_stats.add_tx_by_hash_event(
                    transaction_key.transaction_hash, TransactionStatEventType.TX_REMOVED_FROM_MEMORY,
                    self.network_num, short_id, reason=TxRemovalReason.BLOCK_CLEANUP.value
                )
                self._short_id_to_time_removed[short_id] = time_removed
                removed_sids += 1
                if short_id in self._short_id_to_tx_cache_key:
                    del self._short_id_to_tx_cache_key[short_id]
                    if short_id in self._short_id_to_tx_flag:
                        del self._short_id_to_tx_flag[short_id]
                self._tx_assignment_expire_queue.remove(short_id)
                if self.node.opts.dump_removed_short_ids:
                    self._removed_short_ids.add(short_id)
        else:
            short_ids = None

        if transaction_key.transaction_cache_key in self._tx_cache_key_to_contents:
            self._total_tx_contents_size -= len(self._tx_cache_key_to_contents[transaction_key.transaction_cache_key])
            del self._tx_cache_key_to_contents[transaction_key.transaction_cache_key]
            self._tx_hash_to_time_removed[transaction_key.transaction_cache_key] = time.time()
            removed_txns += 1

        logger.trace("Removed transaction: {}, with {} associated short ids and {} contents.", transaction_key,
                     removed_sids, removed_txns)
        return short_ids

    # pylint: disable=too-many-branches
    def remove_transaction_by_short_id(
        self,
        short_id: int,
        remove_related_short_ids: bool = False,
        force: bool = True,
        removal_reason: TxRemovalReason = TxRemovalReason.UNKNOWN
    ) -> None:
        """
        Clean up short id mapping. Removes transaction contents and mapping if only one short id mapping.
        :param short_id: short id to clean up
        :param remove_related_short_ids: remove all other short id for the same tx
        :param force: when false, cleanup will ignore tx / sids marked with tx flag
        :param removal_reason:
        """
        if short_id in self._short_id_to_tx_cache_key:
            time_removed = time.time()
            transaction_cache_key = self._short_id_to_tx_cache_key[short_id]
            if transaction_cache_key in self._tx_cache_key_to_short_ids:
                short_ids = self._tx_cache_key_to_short_ids[transaction_cache_key]
                short_id_flags = [
                    self._short_id_to_tx_flag.get(sid, TransactionFlag.NO_FLAGS)
                    for sid in short_ids
                ]
                tx_flag = reduce(lambda x, y: x | y, short_id_flags)
                if TransactionFlag.PAID_TX in tx_flag and not force:
                    return
            else:
                short_ids = [short_id]

            if self.node.opts.dump_removed_short_ids:
                self._removed_short_ids.add(short_id)

            del self._short_id_to_tx_cache_key[short_id]
            self._short_id_to_time_removed[short_id] = time_removed
            transaction_hash = self._tx_cache_key_to_hash(transaction_cache_key)
            tx_stats.add_tx_by_hash_event(
                transaction_hash, TransactionStatEventType.TX_REMOVED_FROM_MEMORY,
                self.network_num, short_id, reason=removal_reason.value
            )
            # Only clear mapping and txhash_to_contents if last SID assignment
            if len(short_ids) == 1 or remove_related_short_ids:
                for dup_short_id in short_ids:
                    if dup_short_id != short_id:
                        tx_stats.add_tx_by_hash_event(
                            transaction_hash, TransactionStatEventType.TX_REMOVED_FROM_MEMORY,
                            self.network_num, dup_short_id, reason=removal_reason.value
                        )
                        self._short_id_to_time_removed[dup_short_id] = time_removed
                        if dup_short_id in self._short_id_to_tx_cache_key:
                            del self._short_id_to_tx_cache_key[dup_short_id]
                        self._tx_assignment_expire_queue.remove(dup_short_id)
                        if self.node.opts.dump_removed_short_ids:
                            self._removed_short_ids.add(dup_short_id)

                if transaction_cache_key in self._tx_cache_key_to_contents:
                    self._total_tx_contents_size -= len(self._tx_cache_key_to_contents[transaction_cache_key])
                    del self._tx_cache_key_to_contents[transaction_cache_key]
                    self._tx_hash_to_time_removed[transaction_cache_key] = time_removed

                # Delete short ids from _tx_cache_key_to_short_ids after iterating short_ids.
                # Otherwise extension implementation disposes short_ids list after this line
                del self._tx_cache_key_to_short_ids[transaction_cache_key]
                if not remove_related_short_ids:  # TODO : remove this after creating AbstractTransactionService
                    self._track_seen_transaction(transaction_cache_key)
            else:
                short_ids.remove(short_id)

        self._tx_assignment_expire_queue.remove(short_id)

    def iter_transaction_hashes(self) -> Generator[Sha256Hash, None, None]:
        for tx_cache_key in self._tx_cache_key_to_contents:
            yield self._tx_cache_key_to_hash(tx_cache_key)

    def iter_short_ids_seen_in_block(self) -> Generator[Tuple[Sha256Hash, List[int]], None, None]:
        for block_hash, short_ids in self._short_ids_seen_in_block.items():
            yield block_hash, short_ids

    def iter_transaction_hashes_from_oldest(
        self, newest_time: float = float("inf")
    ) -> Generator[Tuple[Sha256Hash, float], None, None]:
        for short_id, timestamp in self._tx_assignment_expire_queue.queue.items():
            if timestamp > newest_time:
                break

            transaction_hash = self.get_transaction(short_id).hash
            assert transaction_hash is not None
            yield transaction_hash, timestamp

    def thread_safe_iter_transactions_from_oldest(
        self, newest_time: float = float("inf")
    ) -> Generator[Tuple[int, Sha256Hash, float], None, None]:

        tries = 0
        items = None

        while tries < 10:
            try:
                items = list(self._tx_assignment_expire_queue.queue.items())
                break
            except RuntimeError:
                tries += 1

        if items is None:
            raise RuntimeError("The expiration queue changed size during iteration every time")

        logger.debug("Attempted to freeze assignment queue {} times", tries)

        for short_id, timestamp in items:
            if timestamp > newest_time:
                break

            if short_id in self._short_id_to_tx_cache_key:
                try:
                    tx_cache_key = self._short_id_to_tx_cache_key[short_id]
                except KeyError:
                    # ignore, probably a concurrency problem removing transactions from a different thread
                    continue
                tx_hash = self._tx_cache_key_to_hash(tx_cache_key)
                yield short_id, tx_hash, timestamp
            else:
                # ignore, probably a concurrency problem removing transactions from a different thread
                continue

    def on_block_cleaned_up(self, block_hash: Sha256Hash) -> None:
        """
        notify the transaction service about a block transactions cleaned from the pool
        :param block_hash: the block sha
        """
        if block_hash in self._short_ids_seen_in_block:
            del self._short_ids_seen_in_block[block_hash]

    def get_short_ids_seen_in_block_count_info(self) -> Tuple[int, int]:
        """
        :return: tuple of number of blocks and total count of short ids seen in all blocks
        """
        total_short_ids_seen_in_blocks = 0
        for _, seen_short_ids_in_block in self._short_ids_seen_in_block.items():
            total_short_ids_seen_in_blocks += len(seen_short_ids_in_block)

        return self.get_tracked_seen_block_count(), total_short_ids_seen_in_blocks

    def get_snapshot(self, duration: float = 0) -> List[Sha256Hash]:
        if duration > 0:
            snapshot_cache_keys = set()
            remove_expired_sids = []
            snapshot_start_from = time.time() - duration
            for short_id, timestamp in self._tx_assignment_expire_queue.queue.items():
                if timestamp > snapshot_start_from:
                    cache_key = self._short_id_to_tx_cache_key.get(short_id, None)
                    if cache_key is not None:
                        snapshot_cache_keys.add(cache_key)
                    else:
                        logger.trace("Short id: {} does not exist!", short_id)
                        remove_expired_sids.append(short_id)
            for short_id in remove_expired_sids:
                self._tx_assignment_expire_queue.remove(short_id)
            return [self._tx_cache_key_to_hash(tx_cache_key) for tx_cache_key in snapshot_cache_keys]
        else:
            return [self._tx_cache_key_to_hash(tx_cache_key) for tx_cache_key in self._tx_cache_key_to_contents]

    def get_tracked_blocks(self, skip_start: int = 0, skip_end: int = 0) -> Dict[Sha256Hash, int]:
        """
        get a dictionary of tracked seen blocks from start until the reaching a certain number from the end,
        since we want to clear the oldest blocks we already seen.
        :param skip_start: the amount of blocks to skip from start
        :param skip_end: the amount of blocks to skip from the end
        :return: a dictionary of seen blocks mapped to their index relative to the start
        """
        tracked_blocks = {}
        for idx, block in enumerate(self._iter_block_seen_by_time(skip_start, skip_end)):
            tracked_blocks[block] = idx
        return tracked_blocks

    def get_oldest_tracked_block(self, skip_block_count: int) -> List[Sha256Hash]:
        """
        get a list of tracked seen blocks from start until the reaching a certain number from the end,
        since we want to clear the oldest blocks we already seen.
        :param skip_block_count: the amount of blocks to skip from the end (the newest block)
        :return: a list of seen blocks from oldest to latest
        """
        return list(self._iter_block_seen_by_time(0, skip_block_count))

    def get_tracked_seen_block_count(self) -> int:
        return len(self._short_ids_seen_in_block)

    def get_short_id_assign_time(self, short_id: int) -> float:
        if short_id in self._tx_assignment_expire_queue.queue:
            return self._tx_assignment_expire_queue.queue[short_id]
        else:
            logger.warning(log_messages.MISSING_ASSIGN_TIME_FOR_SHORT_ID, short_id)
            return 0.0

    def expire_old_assignments(self) -> float:
        """
        Clean up expired short ids.
        """
        logger.debug("Expiring old short id assignments. Total entries: {}",
                     len(self._tx_assignment_expire_queue))
        self._tx_assignment_expire_queue.remove_expired(
            remove_callback=self.remove_transaction_by_short_id,
            limit=constants.MAX_EXPIRED_TXS_TO_REMOVE,
            force=True,
            removal_reason=TxRemovalReason.EXPIRATION
        )
        logger.debug("Finished cleaning up short ids. Entries remaining: {}",
                     len(self._tx_assignment_expire_queue))
        if len(self._tx_assignment_expire_queue) > 0:
            oldest_tx_timestamp = self._tx_assignment_expire_queue.get_oldest_item_timestamp()
            assert oldest_tx_timestamp is not None
            time_to_expire_oldest = (oldest_tx_timestamp + self.node.opts.sid_expire_time) - time.time()
            return max(time_to_expire_oldest, constants.MIN_CLEAN_UP_EXPIRED_TXS_TASK_INTERVAL_S)
        else:
            self.tx_assign_alarm_scheduled = False
            return 0

    def expire_content_without_sid(self) -> float:
        """
        Clean up content without short ids.
        """
        logger.debug("Expiring tx content for tx without sid. Total entries: {}",
                     len(self.tx_hashes_without_short_id))
        self.tx_hashes_without_short_id.remove_expired(
            remove_callback=self.remove_transaction_by_tx_hash,
            limit=constants.MAX_EXPIRED_TXS_TO_REMOVE,
            force=True,
            assume_no_sid=True
        )
        logger.debug("Finished cleaning up tx. Entries remaining: {}",
                     len(self.tx_hashes_without_short_id))
        if len(self.tx_hashes_without_short_id) > 0:
            oldest_tx_timestamp = self.tx_hashes_without_short_id.get_oldest_item_timestamp()
            assert oldest_tx_timestamp is not None
            time_to_expire_oldest = (oldest_tx_timestamp + constants.TX_CONTENT_NO_SID_EXPIRE_S) - time.time()
            return max(time_to_expire_oldest, constants.MIN_CLEAN_UP_EXPIRED_TXS_TASK_INTERVAL_S)
        else:
            self.tx_content_without_sid_alarm_scheduled = False
            return 0

    def expire_sid_without_content(self) -> float:
        """
        Clean up short ids without content.
        """
        logger.debug("Expiring short ids without content. Total entries: {}",
                     len(self.tx_hashes_without_content))
        self.tx_hashes_without_content.remove_expired(
            remove_callback=self.remove_sid_by_tx_hash,
            limit=constants.MAX_EXPIRED_TXS_TO_REMOVE,
        )
        logger.debug("Finished cleaning up sids. Entries remaining: {}",
                     len(self.tx_hashes_without_content))
        if len(self.tx_hashes_without_content) > 0:
            oldest_tx_timestamp = self.tx_hashes_without_content.get_oldest_item_timestamp()
            assert oldest_tx_timestamp is not None
            time_to_expire_oldest = (oldest_tx_timestamp + constants.TX_CONTENT_NO_SID_EXPIRE_S) - time.time()
            return max(time_to_expire_oldest, constants.MIN_CLEAN_UP_EXPIRED_TXS_TASK_INTERVAL_S)
        else:
            self.tx_without_content_alarm_scheduled = False
            return 0

    def track_seen_short_ids(self, block_hash: Sha256Hash, short_ids: List[int]) -> None:
        """
        Track short ids that has been seen in a routed block.
        That information helps transaction service make a decision when to remove transactions from cache.

        :param block_hash: block hash of the tx short ids
        :param short_ids: transaction short ids
        """

        wrapped_block_hash = wrap_sha256(block_hash)
        self._short_ids_seen_in_block[wrapped_block_hash] = short_ids

        if len(self._short_ids_seen_in_block) >= self._final_tx_confirmations_count:

            # pyre-fixme[28]: Unexpected keyword argument `last`.
            _, final_short_ids = self._short_ids_seen_in_block.popitem(last=False)

            for short_id in final_short_ids:
                self.remove_transaction_by_short_id(short_id, remove_related_short_ids=True, force=True,
                                                    removal_reason=TxRemovalReason.BLOCK_CLEANUP)

        logger_memory_cleanup.statistics(
            {
                "type": "MemoryCleanup",
                "event": "TransactionServiceTrackSeenSummary",
                "data": self.get_cache_state_json(),
                "seen_short_ids_count": len(short_ids),
                "block_hash": repr(block_hash)
            }
        )

    def track_seen_short_ids_delayed(self, block_hash: Sha256Hash, short_ids: List[int]) -> None:
        """
        Schedules alarm task to clean up seen short ids after some delay
        :param block_hash: block hash
        :param short_ids: transaction short ids
        :return:
        """

        self.node.alarm_queue.register_alarm(
            constants.CLEAN_UP_SEEN_SHORT_IDS_DELAY_S,
            self.track_seen_short_ids,
            block_hash,
            short_ids
        )

    def process_tx_sync_message(self, msg: TxServiceSyncTxsMessage) -> List[TxSyncMsgProcessingItem]:
        result_items = []
        txs_content_short_ids = msg.txs_content_short_ids()
        for tx_content_short_ids in txs_content_short_ids:
            transaction_key = self.get_transaction_key(tx_content_short_ids.tx_hash)

            tx_content = tx_content_short_ids.tx_content
            tx_content_len = 0

            if tx_content:
                tx_content_len = len(tx_content)
                self.set_transaction_contents_by_key(transaction_key, tx_content)

            for short_id, _ in zip(
                tx_content_short_ids.short_ids, tx_content_short_ids.short_id_flags
            ):
                self.assign_short_id_by_key(transaction_key, short_id)

            result_item = TxSyncMsgProcessingItem(transaction_key.transaction_hash,
                                                  tx_content_len,
                                                  tx_content_short_ids.short_ids,
                                                  tx_content_short_ids.short_id_flags)
            result_items.append(result_item)

        return result_items

    def log_block_transaction_cleanup_stats(self, block_hash: Sha256Hash, tx_count: int, tx_before_cleanup: int,
                                            tx_after_cleanup: int, short_id_count_before_cleanup: int,
                                            short_id_count_after_cleanup: int):
        logger_memory_cleanup.statistics(
            {
                "type": "BlockTransactionsCleanup",
                "block_hash": repr(block_hash),
                "block_transactions_count": tx_count,
                "tx_hash_to_contents_len_before_cleanup": tx_before_cleanup,
                "tx_hash_to_contents_len_after_cleanup": tx_after_cleanup,
                "short_id_count_before_cleanup": short_id_count_before_cleanup,
                "short_id_count_after_cleanup": short_id_count_after_cleanup,
            }
        )

    def log_tx_service_mem_stats(self, include_data_structure_memory: bool = False) -> None:
        """
        Logs transactions service memory statistics
        """
        if self.node.opts.stats_calculate_actual_size and include_data_structure_memory:
            size_type = memory_utils.SizeType.OBJECT
        else:
            size_type = memory_utils.SizeType.ESTIMATE

        class_name = self.__class__.__name__
        hooks.add_obj_mem_stats(
            class_name,
            self.network_num,
            self._tx_cache_key_to_contents,
            "tx_cache_key_to_contents",
            self.get_collection_mem_stats(
                size_type,
                self._tx_cache_key_to_contents,
                self.ESTIMATED_TX_HASH_ITEM_SIZE * len(self._tx_cache_key_to_contents) + self._total_tx_contents_size
            ),
            object_item_count=len(self._tx_cache_key_to_contents),
            object_type=self.get_object_type(self._tx_cache_key_to_contents),
            size_type=size_type
        )
        hooks.add_obj_mem_stats(
            class_name,
            self.network_num,
            self._short_id_to_tx_cache_key,
            "short_id_to_tx_cache_key",
            self.get_collection_mem_stats(
                size_type,
                self._short_id_to_tx_cache_key,
                self.ESTIMATED_TX_HASH_AND_SHORT_ID_ITEM_SIZE * len(self._short_id_to_tx_cache_key)
            ),
            object_item_count=len(self._short_id_to_tx_cache_key),
            object_type=self.get_object_type(self._short_id_to_tx_cache_key),
            size_type=size_type
        )

        if include_data_structure_memory:
            hooks.add_obj_mem_stats(
                class_name,
                self.network_num,
                self._tx_cache_key_to_short_ids,
                "tx_cache_key_to_short_ids",
                self.get_collection_mem_stats(
                    size_type,
                    self._tx_cache_key_to_short_ids,
                    self.ESTIMATED_TX_HASH_AND_SHORT_ID_ITEM_SIZE * len(self._tx_cache_key_to_short_ids)
                ),
                object_item_count=len(self._tx_cache_key_to_short_ids),
                object_type=self.get_object_type(self._tx_cache_key_to_short_ids),
                size_type=size_type
            )
            hooks.add_obj_mem_stats(
                class_name,
                self.network_num,
                self._short_ids_seen_in_block,
                "short_ids_seen_in_block",
                self.get_collection_mem_stats(
                    size_type,
                    self._short_ids_seen_in_block,
                    self.ESTIMATED_TX_HASH_AND_SHORT_ID_ITEM_SIZE * len(self._short_ids_seen_in_block)
                ),
                object_item_count=len(self._short_ids_seen_in_block),
                object_type=memory_utils.ObjectType.BASE,
                size_type=size_type
            )
            hooks.add_obj_mem_stats(
                class_name,
                self.network_num,
                self._tx_assignment_expire_queue,
                "tx_assignment_expire_queue",
                self.get_collection_mem_stats(
                    size_type,
                    self._tx_assignment_expire_queue,
                    self.ESTIMATED_SHORT_ID_EXPIRATION_ITEM_SIZE * len(self._tx_assignment_expire_queue)
                ),
                object_item_count=len(self._tx_assignment_expire_queue),
                object_type=memory_utils.ObjectType.BASE,
                size_type=size_type
            )

            hooks.add_obj_mem_stats(
                class_name,
                self.network_num,
                self._removed_short_ids,
                "removed_short_ids",
                self.get_collection_mem_stats(
                    size_type,
                    self._removed_short_ids,
                    len(self._removed_short_ids) * constants.SID_LEN
                ),
                object_item_count=len(self._removed_short_ids),
                object_type=memory_utils.ObjectType.BASE,
                size_type=size_type
            )

            hooks.add_obj_mem_stats(
                class_name,
                self.network_num,
                self._tx_hash_to_time_removed,
                "tx_hash_to_time_removed",
                self.get_collection_mem_stats(
                    size_type,
                    self._tx_hash_to_time_removed,
                    len(self._tx_hash_to_time_removed) * SHA256_HASH_LEN
                ),
                object_item_count=len(self._tx_hash_to_time_removed),
                object_type=memory_utils.ObjectType.BASE,
                size_type=size_type
            )

            hooks.add_obj_mem_stats(
                class_name,
                self.network_num,
                self._short_id_to_time_removed,
                "_short_id_to_time_removed",
                self.get_collection_mem_stats(
                    size_type,
                    self._short_id_to_time_removed,
                    len(self._short_id_to_time_removed) * constants.SID_LEN
                ),
                object_item_count=len(self._short_id_to_time_removed),
                object_type=memory_utils.ObjectType.BASE,
                size_type=size_type
            )

            hooks.add_obj_mem_stats(
                class_name,
                self.network_num,
                self.tx_hashes_without_short_id,
                "tx_hash_without_sid",
                self.get_collection_mem_stats(
                    size_type,
                    self.tx_hashes_without_short_id,
                    len(self.tx_hashes_without_short_id) * SHA256_HASH_LEN
                ),
                object_item_count=len(self.tx_hashes_without_short_id),
                object_type=memory_utils.ObjectType.BASE,
                size_type=size_type
            )

            hooks.add_obj_mem_stats(
                class_name,
                self.network_num,
                self.tx_hashes_without_content,
                "tx_hash_sid_without_content",
                self.get_collection_mem_stats(
                    size_type,
                    self.tx_hashes_without_content,
                    len(self.tx_hashes_without_content) * SHA256_HASH_LEN
                ),
                object_item_count=len(self.tx_hashes_without_content),
                object_type=memory_utils.ObjectType.BASE,
                size_type=size_type
            )

    def get_object_type(
        self,
        collection_obj: Any  # pylint: disable=unused-argument
    ):
        return memory_utils.ObjectType.BASE

    def get_aggregate_stats(self) -> Dict[str, Any]:
        """
        Returns dictionary with aggregated statistics of transactions service

        :return: dictionary with aggregated statistics
        """
        oldest_transaction_date = 0
        oldest_transaction_hash = ""

        if len(self._tx_assignment_expire_queue.queue) > 0:
            oldest_transaction_date = self._tx_assignment_expire_queue.get_oldest_item_timestamp()
            oldest_transaction_sid = self._tx_assignment_expire_queue.get_oldest()
            if oldest_transaction_sid in self._short_id_to_tx_cache_key:
                oldest_transaction_hash = self._short_id_to_tx_cache_key[oldest_transaction_sid]

        current_stats = TransactionServiceStats(
            len(self._short_id_to_tx_cache_key),
            len(self._tx_cache_key_to_contents),
            self._total_tx_removed_by_memory_limit,
            self._total_tx_contents_size
        )

        difference = current_stats - self._last_transaction_stats
        self._last_transaction_stats = current_stats

        return {
            "oldest_transaction_date": oldest_transaction_date,
            "oldest_transaction_hash": oldest_transaction_hash,
            "aggregate": current_stats.__dict__,
            "delta": difference.__dict__
        }

    def get_collection_mem_stats(self, size_type: SizeType, collection_obj: Any, estimated_size: int = 0) -> ObjectSize:
        if size_type == SizeType.OBJECT:
            return memory_utils.get_object_size(collection_obj)
        else:
            return ObjectSize(size=estimated_size, flat_size=0, is_actual_size=False)

    def get_cache_state_json(self) -> Dict[str, Any]:
        return {
            "tx_hash_to_short_ids_len": len(self._tx_cache_key_to_short_ids),
            "short_id_to_tx_hash_len": len(self._short_id_to_tx_cache_key),
            "tx_hash_to_contents_len": len(self._tx_cache_key_to_contents),
            "short_ids_seen_in_block_len": len(self._short_ids_seen_in_block),
            "total_tx_contents_size": self._total_tx_contents_size,
            "network_num": self.network_num
        }

    def get_cache_state_str(self) -> str:
        return json_encoder.to_json(self.get_cache_state_json())

    def get_removed_tx_hash_time_and_count(self, tx_hashes: List[Sha256Hash]) -> Tuple[float, int]:
        oldest_removed_tx_hash = 0
        removed_tx_hash_count = 0
        for tx_hash in tx_hashes:
            if tx_hash in self._tx_hash_to_time_removed:
                removed_tx_hash_count += 1
                if oldest_removed_tx_hash < self._tx_hash_to_time_removed[tx_hash]:
                    oldest_removed_tx_hash = self._tx_hash_to_time_removed[tx_hash]
        return oldest_removed_tx_hash, removed_tx_hash_count

    def get_removed_short_id_time_and_count(self, short_ids: List[int]) -> Tuple[float, int]:
        oldest_removed_short_id = 0
        removed_short_id_count = 0
        for short_id in short_ids:
            if short_id in self._short_id_to_time_removed:
                removed_short_id_count += 1
                if oldest_removed_short_id < self._short_id_to_time_removed[short_id]:
                    oldest_removed_short_id = self._short_id_to_time_removed[short_id]
        return oldest_removed_short_id, removed_short_id_count

    def get_transaction_key(
        self,
        transaction_hash: Optional[Sha256Hash],
        transaction_cache_key: Optional[TransactionCacheKeyType] = None
    ) -> TransactionKey:
        if transaction_hash is None:
            assert transaction_cache_key is not None
            transaction_hash = self._tx_cache_key_to_hash(transaction_cache_key)

        transaction_hash = wrap_sha256(transaction_hash)
        return TransactionKey(
            transaction_hash,
            transaction_cache_key,
            _lazy_transaction_cache_key=iter((self._tx_hash_to_cache_key(transaction_hash)) for _ in iter((None,)))
        )

    def _dump_removed_short_ids(self) -> int:
        if self._removed_short_ids:
            with open(f"{self.node.opts.dump_removed_short_ids_path}/{int(time.time())}", "w") as file_handle:
                file_handle.write(str(self._removed_short_ids))
            self._removed_short_ids.clear()
        return constants.DUMP_REMOVED_SHORT_IDS_INTERVAL_S

    def _memory_limit_clean_up(self) -> None:
        """
        Removes oldest transactions if total bytes consumed by transaction contents exceed memory limit
        """
        if self._total_tx_contents_size <= self._tx_content_memory_limit:
            return

        logger.trace("Transaction service exceeds memory limit for transaction contents. Limit: {}. Current size: {}.",
                     self._tx_content_memory_limit, self._total_tx_contents_size)
        removed_tx_count = 0

        while self._is_exceeding_memory_limit() and self._tx_assignment_expire_queue:
            self._tx_assignment_expire_queue.remove_oldest(
                remove_callback=self.remove_transaction_by_short_id,
                removal_reason=TxRemovalReason.MEMORY_LIMIT
            )
            removed_tx_count += 1
        if self._is_exceeding_memory_limit() and not self._tx_assignment_expire_queue:
            logger.warning(log_messages.SID_MEMORY_MANAGEMENT_FAILURE, self.get_cache_state_json())
            removed_tx_count += len(self._tx_cache_key_to_contents)
            self.clear()

        self._total_tx_removed_by_memory_limit += removed_tx_count
        logger.trace("Removed {} oldest transactions from transaction service cache. Size after clean up: {}",
                     removed_tx_count, self._total_tx_contents_size)

    def _get_final_tx_confirmations_count(self) -> int:
        """
        Returns configuration value of number of block confirmations required before transaction can be removed
        """
        if self.network:
            return self.network.final_tx_confirmations_count

        logger.warning(
            log_messages.UNABLE_TO_DETERMINE_TX_FINAL_CONFIRMATIONS_COUNT,
            self.network_num, self.DEFAULT_FINAL_TX_CONFIRMATIONS_COUNT
        )

        return self.DEFAULT_FINAL_TX_CONFIRMATIONS_COUNT

    def _get_removed_transactions_history_expiration_time_s(self) -> int:
        """
        Returns configuration value of expiration time for transaction hashes removed from cache
        """
        if self.network:
            return self.network.removed_transactions_history_expiration_s

        logger.warning(
            log_messages.UNABLE_TO_DETERMINE_TX_EXPIRATION_TIME,
            self.network_num, constants.REMOVED_TRANSACTIONS_HISTORY_EXPIRATION_S
        )

        return constants.REMOVED_TRANSACTIONS_HISTORY_EXPIRATION_S

    def _get_tx_contents_memory_limit(self) -> int:
        """
        Returns configuration value for memory limit for total transaction contents
        """
        if self.node.opts.transaction_pool_memory_limit is not None:
            # convert MB to bytes
            return int(self.node.opts.transaction_pool_memory_limit * 1024 * 1024)

        if self.network:
            if self.network.tx_contents_memory_limit_bytes is None:
                logger.warning(
                    log_messages.TX_CACHE_SIZE_LIMIT_NOT_CONFIGURED,
                    self.network_num, constants.DEFAULT_TX_CACHE_MEMORY_LIMIT_BYTES
                )
                return constants.DEFAULT_TX_CACHE_MEMORY_LIMIT_BYTES
            else:
                return self.network.tx_contents_memory_limit_bytes

        logger.warning(
            log_messages.UNABLE_TO_DETERMINE_TX_MEMORY_LIMIT,
            self.network_num, constants.DEFAULT_TX_CACHE_MEMORY_LIMIT_BYTES
        )
        return constants.DEFAULT_TX_CACHE_MEMORY_LIMIT_BYTES

    def _tx_hash_to_cache_key(self, transaction_hash: Union[Sha256Hash, bytes, bytearray, memoryview, str]) -> str:

        if isinstance(transaction_hash, Sha256Hash):
            return convert.bytes_to_hex(transaction_hash.binary)

        if isinstance(transaction_hash, (bytes, bytearray, memoryview)):
            return convert.bytes_to_hex(transaction_hash)

        # pyre-fixme[25]: Assertion will always fail.
        if isinstance(transaction_hash, str):
            return transaction_hash

        raise ValueError("Attempted to find cache entry with incorrect key type")

    def _wrap_sha256(self, transaction_hash: Union[bytes, bytearray, memoryview, Sha256Hash]) -> Sha256Hash:
        if isinstance(transaction_hash, Sha256Hash):
            return transaction_hash

        # pyre-fixme[25]: Assertion will always fail.
        if isinstance(transaction_hash, (bytes, bytearray, memoryview)):
            return Sha256Hash(binary=transaction_hash)

        return Sha256Hash(binary=convert.hex_to_bytes(transaction_hash))

    def _tx_cache_key_to_hash(self, transaction_cache_key: TransactionCacheKeyType) -> Sha256Hash:
        if isinstance(transaction_cache_key, Sha256Hash):
            return transaction_cache_key

        if isinstance(transaction_cache_key, (bytes, bytearray, memoryview)):
            return Sha256Hash(transaction_cache_key)

        # pyre-fixme[6]:
        return Sha256Hash(convert.hex_to_bytes(transaction_cache_key))

    def _track_seen_transaction(self, transaction_cache_key):
        pass

    def _iter_block_seen_by_time(self, skip_start: int, skip_end: int) -> Iterator[Sha256Hash]:
        end = len(self._short_ids_seen_in_block) - skip_end
        for idx, block_hash in enumerate(self._short_ids_seen_in_block.keys()):
            if skip_start <= idx <= end:
                yield block_hash

    def _is_exceeding_memory_limit(self) -> bool:
        return self._total_tx_contents_size > self._tx_content_memory_limit

    def clear(self) -> None:
        self._tx_cache_key_to_contents.clear()
        self._tx_cache_key_to_short_ids.clear()
        self._short_id_to_tx_cache_key.clear()
        self._short_ids_seen_in_block.clear()
        self._short_id_to_tx_flag.clear()
        self.tx_hashes_without_content.clear()
        self.tx_hashes_without_short_id.clear()
        self._tx_assignment_expire_queue.clear()
        self._total_tx_contents_size = 0

    # TODO: remove this unused function
    def _log_transaction_service_histogram(self) -> None:
        """
        logs a histogram of the tracked transactions age,
        buckets are named as according to the age in hours from now
        bucket is named according the the range start

        """
        bucket_count = constants.TRANSACTION_SERVICE_TRANSACTIONS_HISTOGRAM_BUCKETS
        cell_size = self.node.opts.sid_expire_time / bucket_count
        cell_size_hours = cell_size / (60 * 60)
        histogram = Counter()
        current_time = time.time()
        timestamps = self._tx_assignment_expire_queue.queue.values()
        for timestamp in timestamps:
            histogram[int((timestamp // cell_size) * cell_size)] += 1
        logger_tx_histogram.statistics(
            {"type": "TransactionHistogram",
             "data": {datetime.fromtimestamp(k): v for (k, v) in histogram.items()},
             "start_time": datetime.utcnow(),
             "duration": time.time() - current_time,
             "cell_size_s": cell_size,
             "cell_size_h": cell_size_hours,
             "network_num": self.network_num
             }
        )
        return constants.TRANSACTION_SERVICE_LOG_TRANSACTIONS_INTERVAL_S

    def _cleanup_removed_transactions_history(self) -> int:
        """
        Removes expired items from the queue
        """
        logger.trace(
            "Starting to cleanup transaction cache history for network "
            "number: {}.",
            self.network_num
        )

        current_time = time.time()
        tx_hash_history_len_before = len(self._tx_hash_to_time_removed)
        if self._tx_hash_to_time_removed:
            oldest_tx = next(iter(self._tx_hash_to_time_removed))
            while (self._tx_hash_to_time_removed and
                   ((current_time - self._tx_hash_to_time_removed[oldest_tx] > self._removed_txs_expiration_time_s) or
                    len(self._tx_hash_to_time_removed) > constants.REMOVED_TRANSACTIONS_HISTORY_LENGTH_LIMIT)):
                del self._tx_hash_to_time_removed[oldest_tx]
                if not self._tx_hash_to_time_removed:
                    break
                oldest_tx = next(iter(self._tx_hash_to_time_removed))
        tx_hash_history_len_after = len(self._tx_hash_to_time_removed)

        short_id_history_len_before = len(self._short_id_to_time_removed)
        if self._short_id_to_time_removed:
            oldest_short_id = next(iter(self._short_id_to_time_removed))
            while (self._short_id_to_time_removed and
                   ((current_time - self._short_id_to_time_removed[oldest_short_id] >
                     self._removed_txs_expiration_time_s) or
                    len(self._short_id_to_time_removed) > constants.REMOVED_TRANSACTIONS_HISTORY_LENGTH_LIMIT)):
                del self._short_id_to_time_removed[oldest_short_id]
                if not self._short_id_to_time_removed:
                    break
                oldest_short_id = next(iter(self._short_id_to_time_removed))
        short_id_history_len_after = len(self._short_id_to_time_removed)

        logger.trace(
            "Finished cleanup transaction cache history. "
            "tx_hash size before: {}, size after: {}."
            "short_id size before: {}, size after: {}.",
            tx_hash_history_len_before,
            tx_hash_history_len_after,
            short_id_history_len_before,
            short_id_history_len_after
        )

        return constants.REMOVED_TRANSACTIONS_HISTORY_CLEANUP_INTERVAL_S



from typing import TypeVar, Generic, Optional, Dict

from bxcommon.utils.alarm_queue import AlarmQueue
from bxcommon.utils.expiration_queue import ExpirationQueue

KT = TypeVar("KT")
VT = TypeVar("VT")


class ExpiringDict(Generic[KT, VT]):
    """
    Dictionary with expiration time.
    """

    contents: Dict[KT, VT]
    _alarm_queue: AlarmQueue
    _expiration_queue: ExpirationQueue[KT]
    _expiration_time: int
    _name: str

    def __init__(self, alarm_queue: AlarmQueue, expiration_time_s: int, name: str) -> None:
        self.contents = {}
        self._alarm_queue = alarm_queue
        self._expiration_queue = ExpirationQueue(expiration_time_s)
        self._expiration_time = expiration_time_s
        self._name = name

    def __contains__(self, item: KT):
        return item in self.contents

    def __setitem__(self, key: KT, value: VT):
        if key in self.contents:
            self.contents[key] = value
        else:
            self.add(key, value)

    def __delitem__(self, key: KT):
        del self.contents[key]
        self._expiration_queue.remove(key)

    def __getitem__(self, item: KT) -> VT:
        return self.contents[item]

    def add(self, key: KT, value: VT) -> None:
        self.contents[key] = value
        self._expiration_queue.add(key)
        self._alarm_queue.register_approx_alarm(
            self._expiration_time * 2,
            self._expiration_time,
            self.cleanup,
            alarm_name=f"ExpiringDict[{self._name}]#cleanup"
        )

    def cleanup(self) -> float:
        self._expiration_queue.remove_expired(remove_callback=self.remove_item)
        return 0

    def remove_item(self, key: KT) -> Optional[VT]:
        if key in self.contents:
            return self.contents.pop(key)
        else:
            return None



import struct
from typing import List, Union, NamedTuple

from bxcommon.constants import UL_INT_SIZE_IN_BYTES
from bxcommon.utils.crypto import SHA256_HASH_LEN
from bxcommon.utils.object_hash import Sha256Hash


class BlockShortIds(NamedTuple):
    block_hash: Sha256Hash
    short_ids: List[int]


def get_serialized_blocks_short_ids_bytes_len(blocks_short_ids: List[BlockShortIds]) -> int:
    """
    Calculates length of list of blocks short ids serialized into bytes

    :param blocks_short_ids: list of block short ids
    :return: length of serialized bytes
    """

    return len(blocks_short_ids) * SHA256_HASH_LEN + \
           len(blocks_short_ids) * UL_INT_SIZE_IN_BYTES + \
           sum([len(block_short_ids.short_ids) for block_short_ids in blocks_short_ids]) * UL_INT_SIZE_IN_BYTES


def serialize_blocks_short_ids_into_bytes(block_short_ids: List[BlockShortIds]) -> bytearray:
    """
    Serializes list of block short ids into bytes

    :param block_short_ids: list of tuple block short ids
    :return: bytearray with serialized bytes
    """

    buffer = bytearray(get_serialized_blocks_short_ids_bytes_len(block_short_ids))
    off = 0
    for block_short_id in block_short_ids:
        buffer[off:off + SHA256_HASH_LEN] = block_short_id.block_hash.binary
        off += SHA256_HASH_LEN
        struct.pack_into("<L", buffer, off, len(block_short_id.short_ids))
        off += UL_INT_SIZE_IN_BYTES
        for short_id in block_short_id.short_ids:
            struct.pack_into("<L", buffer, off, short_id)
            off += UL_INT_SIZE_IN_BYTES

    return buffer


def deserialize_blocks_short_ids_from_buffer(
    buffer: Union[bytearray, memoryview], offset: int, block_count: int
) -> List[BlockShortIds]:
    """
    Deserialize list of block short ids from buffer

    :param buffer: buffer containing serialized short ids bytes
    :param offset: offset in a buffer where serialized short ids begin
    :param block_count: how many blocks to deserialize
    :return: list of block short ids
    """
    block_short_ids: List[BlockShortIds] = []

    for _ in range(block_count):
        block_hash = Sha256Hash(buffer[offset:offset + SHA256_HASH_LEN])
        offset += SHA256_HASH_LEN
        short_ids_count, = struct.unpack_from("<L", buffer, offset)
        offset += UL_INT_SIZE_IN_BYTES
        short_ids = []
        for _ in range(short_ids_count):
            short_id, = struct.unpack_from("<L", buffer, offset)
            short_ids.append(short_id)
            offset += UL_INT_SIZE_IN_BYTES

        block_short_ids.append(BlockShortIds(block_hash, short_ids))

    return block_short_ids



import argparse
import os
import sys
from argparse import ArgumentParser
from typing import List, Optional, Union

from bxcommon import constants
from bxcommon.constants import ALL_NETWORK_NUM
from bxcommon.models.blockchain_network_model import BlockchainNetworkModel
from bxcommon.models.node_type import NodeType
from bxcommon.rpc import rpc_constants
from bxcommon.services import http_service
from bxcommon.services import sdn_http_service
from bxcommon.utils import config, ip_resolver, convert, node_cache
from bxcommon.utils.node_start_args import NodeStartArgs
from bxutils import constants as utils_constants
from bxutils import log_messages
from bxutils import logging
from bxutils.logging import log_config
from bxutils.logging.log_format import LogFormat
from bxutils.logging.log_level import LogLevel

logger = logging.get_logger(__name__)


def get_argument_parser() -> argparse.ArgumentParser:
    arg_parser = argparse.ArgumentParser(add_help=False)

    arg_parser.add_argument("--external-ip", help="External network ip of this node",
                            type=ip_resolver.blocking_resolve_ip)
    arg_parser.add_argument("--external-port", help="External network port to listen on", type=int,
                            default=config.get_env_default(NodeStartArgs.EXTERNAL_PORT))
    arg_parser.add_argument("--non-ssl-port", help="External network port for non SSL nodes to listen on", type=int,
                            default=config.get_env_default(NodeStartArgs.NON_SSL_PORT))
    arg_parser.add_argument("--continent", help="The continent of this node", type=str,
                            choices=["NA", "SA", "EU", "AS", "AF", "OC", "AN"])
    arg_parser.add_argument("--country", help="The country of this node.", type=str)
    arg_parser.add_argument("--region", help="The region of this node.", type=str)
    arg_parser.add_argument("--hostname", help="Hostname the node is running on", type=str,
                            default=constants.HOSTNAME)
    arg_parser.add_argument("--sdn-url", help="IP or dns of the bloxroute SDN", type=str,
                            default=config.get_env_default(NodeStartArgs.SDN_ROOT_URL))
    arg_parser.add_argument(
        "--node-id",
        help="(TEST ONLY) Set the node_id for using in testing."
    )
    arg_parser.add_argument("--transaction-pool-memory-limit",
                            help="Maximum size of transactions to keep in memory pool (MB)",
                            type=int)

    arg_parser.add_argument("--info-stats-interval",
                            help="Frequency of info statistics logs in seconds",
                            type=int,
                            default=constants.INFO_STATS_INTERVAL_S)

    arg_parser.add_argument("--throughput-stats-interval",
                            help="Frequency of throughput statistics logs in seconds",
                            type=int,
                            default=constants.THROUGHPUT_STATS_INTERVAL_S)

    arg_parser.add_argument("--memory-stats-interval",
                            help="Frequency of memory statistics logs in seconds",
                            type=int,
                            default=constants.MEMORY_STATS_INTERVAL_S)

    arg_parser.add_argument("--dump-detailed-report-at-memory-usage",
                            help="Total memory usage of application when detailed memory report "
                                 "should be dumped to log (MB)",
                            type=int,
                            default=(2 * 1024))
    arg_parser.add_argument("--dump-removed-short-ids",
                            help="Dump removed short ids to a file at a fixed interval",
                            type=convert.str_to_bool,
                            default=False)
    arg_parser.add_argument("--dump-removed-short-ids-path",
                            help="Folder to dump removed short ids to",
                            type=str,
                            default=constants.DUMP_REMOVED_SHORT_IDS_PATH)
    arg_parser.add_argument("--enable-buffered-send", help="Enables buffering of sent byte to improve performance",
                            type=convert.str_to_bool, default=False)
    arg_parser.add_argument("--track-detailed-sent-messages", help="Enables tracking of messages written on socket",
                            type=convert.str_to_bool, default=False)
    arg_parser.add_argument(
        "--use-extensions",
        help="If true than the node will use the extension module for "
             "some tasks like block compression (default: {0})".format(constants.USE_EXTENSION_MODULES),
        default=constants.USE_EXTENSION_MODULES,
        type=convert.str_to_bool
    )

    arg_parser.add_argument(
        "--import-extensions",
        help="If true than the node will import all C++ extensions dependencies on start up",
        default=False,
        type=convert.str_to_bool
    )
    arg_parser.add_argument(
        "--thread-pool-parallelism-degree",
        help="The degree of parallelism to use when running task on a "
        f"concurrent thread pool (default: {constants.DEFAULT_THREAD_POOL_PARALLELISM_DEGREE})",
        default=constants.DEFAULT_THREAD_POOL_PARALLELISM_DEGREE,
        type=config.get_thread_pool_parallelism_degree
    )
    arg_parser.add_argument(
        "--tx-mem-pool-bucket-size",
        help="The size of each bucket of the transaction mem pool. "
             "In order to efficiently iterate the mem pool concurrently, it is being split into buckets. "
        f"(default: {constants.DEFAULT_TX_MEM_POOL_BUCKET_SIZE})",
        default=constants.DEFAULT_TX_MEM_POOL_BUCKET_SIZE,
        type=int
    )
    arg_parser.add_argument(
        "--sync-tx-service",
        help="sync tx service in node",
        type=convert.str_to_bool,
        default=True
    )
    arg_parser.add_argument(
        "--block-compression-debug",
        help="Enable debug messages with details of the compressed blocks",
        type=convert.str_to_bool,
        default=True
    )
    arg_parser.add_argument(
        "--enable-tcp-quickack",
        help="Enable TCP_QUICKACK so that ACK's are not delayed",
        type=convert.str_to_bool,
        default=True
    )

    add_argument_parser_logging(arg_parser)
    add_argument_parser_common(arg_parser)
    return arg_parser


def add_argument_parser_rpc(
    arg_parser: ArgumentParser,
    default_rpc_host=rpc_constants.DEFAULT_RPC_HOST,
    default_rpc_port=rpc_constants.DEFAULT_RPC_PORT
):
    arg_parser.add_argument(
        "--rpc",
        help="Start a HTTP(S) JSON-RPC server",
        type=convert.str_to_bool,
        default=True
    )
    arg_parser.add_argument(
        "--rpc-host",
        help="The node RPC host (default: {}).".format(rpc_constants.DEFAULT_RPC_HOST),
        type=str,
        default=default_rpc_host
    )
    arg_parser.add_argument(
        "--rpc-port",
        help="The node RPC port (default: {}).".format(rpc_constants.DEFAULT_RPC_PORT),
        type=int,
        default=default_rpc_port,
    )
    arg_parser.add_argument(
        "--rpc-use-ssl",
        help="Use secured communication (HTTPS, WSS)",
        type=convert.str_to_bool,
        default=False,
    )
    arg_parser.add_argument(
        "--rpc-ssl-base-url",
        help="The base url for ca, cert, and key used by the RPC server",
        type=str,
        default=rpc_constants.DEFAULT_RPC_BASE_SSL_URL,
    )
    arg_parser.add_argument(
        "--rpc-user",
        help=f"The node RPC server user (default: {rpc_constants.DEFAULT_RPC_USER})",
        type=str,
        default=rpc_constants.DEFAULT_RPC_USER,
    )
    arg_parser.add_argument(
        "--rpc-password",
        help=f"The node RPC server password (default: {rpc_constants.DEFAULT_RPC_PASSWORD})",
        type=str,
        default=rpc_constants.DEFAULT_RPC_PASSWORD,
    )


def add_argument_parser_logging(
    arg_parser: ArgumentParser,
    default_log_level: Union[LogLevel, str] = utils_constants.DEFAULT_LOG_LEVEL
):
    default_log_level = os.environ.get("LOG_LEVEL", default_log_level)
    default_log_format = os.environ.get("LOG_FORMAT", utils_constants.DEFAULT_LOG_FORMAT)
    default_log_fluentd_enable = os.environ.get("LOG_FLUENTD_ENABLE", False)
    default_log_fluentd_host = os.environ.get("LOG_FLUENTD_HOST", utils_constants.FLUENTD_HOST)
    default_fluentd_logger_max_queue_size = os.environ.get(
        "FLUENTD_LOGGER_MAX_QUEUE_SIZE", utils_constants.FLUENTD_LOGGER_MAX_QUEUE_SIZE
    )
    default_log_level_overrides = os.environ.get("LOG_LEVEL_OVERRIDES", {})
    default_log_level_fluentd = os.environ.get("LOG_LEVEL_FLUENTD", LogLevel.NOTSET)
    default_log_level_stdout = os.environ.get("LOG_LEVEL_STDOUT", LogLevel.NOTSET)

    arg_parser.add_argument(
        "--log-level",
        help="set log level",
        # pyre-fixme[16]: `LogLevel` has no attribute `__getattr__`.
        type=LogLevel.__getattr__,
        choices=list(LogLevel),
        default=default_log_level
    )
    arg_parser.add_argument(
        "--log-format",
        help="set log format",
        # pyre-fixme[16]: `LogFormat` has no attribute `__getattr__`.
        type=LogFormat.__getattr__,
        choices=list(LogFormat),
        default=default_log_format
    )
    arg_parser.add_argument(
        "--log-fluentd-enable",
        help="enable logging directly to fluentd",
        type=convert.str_to_bool,
        default=default_log_fluentd_enable
    )
    arg_parser.add_argument(
        "--log-fluentd-host",
        help="fluentd instance address provide, hostname:port",
        type=str,
        default=default_log_fluentd_host
    )
    arg_parser.add_argument(
        "--log-fluentd-queue-size",
        help="fluentd queue size",
        type=int,
        default=default_fluentd_logger_max_queue_size
    )
    arg_parser.add_argument(
        "--log-flush-immediately",
        help="Enables immediate flush for logs",
        type=convert.str_to_bool,
        default=False
    )
    arg_parser.add_argument(
        "--log-level-overrides",
        help="override log level for namespace stats=INFO,bxcommon.connections=WARNING",
        default=default_log_level_overrides,
        type=log_config.str_to_log_options
    )
    arg_parser.add_argument(
        "--log-level-fluentd",
        help="The fluentd handler log level",
        type=LogLevel.__getattr__,
        choices=list(LogLevel),
        default=default_log_level_fluentd
    )
    arg_parser.add_argument(
        "--log-level-stdout",
        help="The stdout handler log level",
        type=LogLevel.__getattr__,
        choices=list(LogLevel),
        default=default_log_level_stdout
    )
    arg_parser.add_argument(
        "--transaction-validation",
        help="Enable transaction validation process",
        type=convert.str_to_bool,
        default=True
    )


def add_argument_parser_common(arg_parser: ArgumentParser):
    arg_parser.add_argument(
        "--data-dir",
        help="Path to store configuration, state and log files",
        default=config.get_default_data_path()
    )
    arg_parser.add_argument(
        "--ca-cert-url",
        help="The URL for retrieving BDN ca certificate data (default: {})".format(
            config.get_env_default(NodeStartArgs.CA_CERT_URL)
        ),
        default=config.get_env_default(NodeStartArgs.CA_CERT_URL),
        type=str
    )
    arg_parser.add_argument(
        "--private-ssl-base-url",
        help="The base URL for retrieving specific certificate data (default: {})".format(
            config.get_env_default(NodeStartArgs.PRIVATE_SSL_BASE_URL)
        ),
        default=config.get_env_default(NodeStartArgs.PRIVATE_SSL_BASE_URL),
        type=str
    )


def add_feed_source_arguments(arg_parser: argparse.ArgumentParser) -> None:
    arg_parser.add_argument("--source-feed-ip", help="Source feed ip", type=str)
    arg_parser.add_argument("--source-feed-port", help="Source feed port", type=int)
    arg_parser.add_argument("--source-feed-rpc-user", help="Source feed user", type=str)
    arg_parser.add_argument("--source-feed-rpc-password", help="Source feed pass", type=str)


def parse_arguments(arg_parser: argparse.ArgumentParser, args: Optional[List[str]] = None) -> argparse.Namespace:
    opts, _unknown = arg_parser.parse_known_args(args)
    if not opts.external_ip:
        opts.external_ip = ip_resolver.get_node_public_ip()
    assert opts.external_ip is not None
    opts.external_ip = ip_resolver.blocking_resolve_ip(opts.external_ip)
    http_service.set_root_url(opts.sdn_url)
    config.append_manifest_args(opts.__dict__)
    return opts


def parse_blockchain_opts(opts, node_type: NodeType):
    """
    Get the blockchain network info from the SDN and set the default values for the blockchain cli params if they were
    not passed in the args.

    :param opts: argument list
    :param node_type:
    """
    opts_dict = opts.__dict__

    if node_type in NodeType.RELAY or node_type in NodeType.BLOXROUTE_CLOUD_API:
        opts_dict["blockchain_network_num"] = ALL_NETWORK_NUM
        return

    network_info = _get_blockchain_network_info(opts)

    for key, value in opts_dict.items():
        if value is None and network_info.default_attributes is not None and key in network_info.default_attributes:
            opts_dict[key] = network_info.default_attributes[key]

    opts_dict["blockchain_network_num"] = network_info.network_num
    opts_dict["blockchain_block_interval"] = network_info.block_interval
    opts_dict["blockchain_ignore_block_interval_count"] = network_info.ignore_block_interval_count
    opts_dict["blockchain_block_recovery_timeout_s"] = network_info.block_recovery_timeout_s
    opts_dict["blockchain_block_hold_timeout_s"] = network_info.block_hold_timeout_s
    opts_dict["enable_network_content_logs"] = network_info.enable_network_content_logs

    if opts.enable_block_compression is None:
        opts_dict["enable_block_compression"] = network_info.enable_block_compression


def _set_blockchain_networks_from_cache(opts):
    cache_info = node_cache.read(opts)
    if cache_info:
        opts.blockchain_networks = cache_info.blockchain_networks
    if not opts.blockchain_networks:
        logger.warning(log_messages.EMPTY_BLOCKCHAIN_NETWORK_CACHE)


def set_blockchain_networks_info(opts):
    opts.blockchain_networks = sdn_http_service.fetch_blockchain_networks()
    if not opts.blockchain_networks:
        logger.warning(log_messages.EMPTY_BLOCKCHAIN_NETWORK_LIST)
        _set_blockchain_networks_from_cache(opts)


def _get_blockchain_network_info(opts) -> BlockchainNetworkModel:
    """
    Retrieves the blockchain network info from the SDN based on blockchain-protocol
    and blockchain-network cli arguments.

    :param opts: argument list
    """

    for blockchain_network in opts.blockchain_networks.values():
        if blockchain_network.protocol.lower() == opts.blockchain_protocol.lower() and \
                blockchain_network.network.lower() == opts.blockchain_network.lower():
            return blockchain_network

    blockchain_networks = opts.blockchain_networks
    cache_network_info = node_cache.read(opts)
    if not blockchain_networks and cache_network_info:
        blockchain_networks = cache_network_info.blockchain_networks

    if blockchain_networks:
        all_networks_names = "\n".join(
            map(lambda n: "{} - {}".format(n.protocol, n.network), blockchain_networks.values())
        )
        error_msg = "Network number does not exist for blockchain protocol {} and network {}.\nValid options:\n{}" \
            .format(opts.blockchain_protocol, opts.blockchain_network, all_networks_names)
        logger.fatal(error_msg, exc_info=False)
    else:
        logger.fatal("Could not reach the SDN to fetch network information. Check that {} is the actual address "
                     "you are trying to reach.", opts.sdn_url, exc_info=False)
    sys.exit(1)



from typing import Optional

from bxcommon import constants
from bxcommon.messages.bloxroute.bloxroute_message_type import BloxrouteMessageType
from bxcommon.messages.bloxroute.keep_alive_message import KeepAliveMessage
from bxcommon.messages.bloxroute.abstract_bloxroute_message import AbstractBloxrouteMessage
from bxcommon.messages.bloxroute.protocol_version import PROTOCOL_VERSION
from bxcommon.utils.message_buffer_builder import PayloadElement, PayloadBlock


class PongMessage(KeepAliveMessage):
    MESSAGE_TYPE = BloxrouteMessageType.PONG

    PONG_MESSAGE_BLOCK = PayloadBlock(
        AbstractBloxrouteMessage.HEADER_LENGTH,
        "ResponseMessage",
        PROTOCOL_VERSION,
        PayloadElement(name="nonce", structure="<Q", decode=lambda x: x or None),
        PayloadElement(name="timestamp", structure="<Q", decode=lambda x: x or None),
    )
    PONG_MESSAGE_LENGTH = PONG_MESSAGE_BLOCK.size + constants.CONTROL_FLAGS_LEN

    def __init__(
        self, nonce: Optional[int] = None, timestamp: Optional[int] = None, buf: Optional[bytearray] = None
    ) -> None:
        if buf is None:
            buf = bytearray(self.HEADER_LENGTH + self.PONG_MESSAGE_LENGTH)

        buf = self.PONG_MESSAGE_BLOCK.build(buf, timestamp=timestamp)

        self._timestamp: Optional[int] = None
        super(PongMessage, self).__init__(
            msg_type=self.MESSAGE_TYPE, nonce=nonce, buf=buf, payload_length=self.PONG_MESSAGE_LENGTH
        )

    def __unpack(self) -> None:
        contents = self.PONG_MESSAGE_BLOCK.read(self._memoryview)
        self._nonce = contents.get("nonce")
        self._timestamp = contents.get("timestamp")

    def timestamp(self) -> Optional[int]:
        if self._timestamp is None:
            self.__unpack()
        return self._timestamp

    def __repr__(self) -> str:
        return "PongMessage<nonce: {},timestamp: {}>".format(self.nonce(), self.timestamp())



from typing import NamedTuple, Optional

from bxutils.ssl.data.ssl_file_info import SSLFileInfo


class SSLCertificateInfo(NamedTuple):
    cert_file_info: SSLFileInfo
    key_file_info: Optional[SSLFileInfo] = None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} <" \
            f"cert_file_info = {self.cert_file_info}, " \
            f"key_file_info = {self.key_file_info}" \
            f">"



from bxcommon.utils.blockchain_utils.ont import ont_common_constants

ONT_MAGIC_NUMBERS = {
    "mainnet": 0x8C77AB60,
    "polaris": 0x2D8829DF
}

STARTUP_CAP = bytes(32)
STARTUP_RELAY_STATE = True
STARTUP_IS_CONSENSUS = True

ONT_HDR_COMMON_OFF = 24
ONT_HEADER_MINUS_CHECKSUM = 20

ONT_HASH_LEN = ont_common_constants.ONT_HASH_LEN
ONT_BOOKKEEPER_LEN = 33
ONT_BOOKKEEPER_AND_VARINT_LEN = 42
ONT_NODE_ADDR_LEN = 44
ONT_VARINT_MAX_LEN = 9
ONT_IP_PORT_SIZE = 2
ONT_IP_ADDR_SIZE = 16

ONT_LONG_LONG_LEN = 8
ONT_INT_LEN = 4
ONT_SHORT_LEN = 2
ONT_CHAR_LEN = 1

ONT_MSG_CHECKSUM_LEN = 4
ONT_ADDR_TIME_AND_SERV_LEN = 16
ONT_ADDR_PORTS_AND_ID_LEN = 12
ONT_BLOCK_TIME_HEIGHT_CONS_DATA_LEN = 16
ONT_BLOCK_NEXT_BOOKKEEPER_LEN = 20
ONT_DATA_MSG_LEN = 65
ONT_GET_DATA_MSG_LEN = 33
ONT_VERSION_VER_SERV_TIME_LEN = 20
ONT_VERSION_PORTS_LEN = 6
ONT_VERSION_LOAD_LEN = 51
ONT_VERSION_MSG_LEN = 101

BLOCK_PROPOSAL_CONSENSUS_MESSAGE_TYPE = 0

ONT_NODE_SERVICES = 1

ONT_HELLO_MESSAGES = [b"version", b"verack"]

ONT_SHORT_ID_INDICATOR = 0xFF
ONT_SHORT_ID_INDICATOR_AS_BYTEARRAY = bytearray([ONT_SHORT_ID_INDICATOR])
ONT_SHORT_ID_INDICATOR_LENGTH = 1

ONT_PING_INTERVAL_S = 3
ONT_MAX_QUEUED_BLOCKS = 20

ONT_TX_DEPLOY_TYPE_INDICATOR = 0xD0
ONT_TX_DEPLOY_TYPE_INDICATOR_AS_BYTEARRAY = bytearray([ONT_TX_DEPLOY_TYPE_INDICATOR])
ONT_TX_INVOKE_TYPE_INDICATOR = 0xD1
ONT_TX_INVOKE_TYPE_INDICATOR_AS_BYTEARRAY = bytearray([ONT_TX_INVOKE_TYPE_INDICATOR])

ONT_DEFAULT_BLOCK_SIZE = 621000
ONT_MINIMAL_SUB_TASK_TX_COUNT = 2500

BLOCK_CONFIRMATION_REQUEST_INTERVALS = 5
BLOCK_CONFIRMATION_REQUEST_CACHE_INTERVAL_S = 30 * 60

TRACKED_BLOCK_CLEANUP_INTERVAL_S = 15
BLOCK_CLEANUP_NODE_BLOCK_LIST_POLL_INTERVAL_S = 0



import struct
from typing import List, Dict

from bxcommon.constants import UL_INT_SIZE_IN_BYTES
from bxcommon.messages.abstract_block_message import AbstractBlockMessage
from bxcommon.utils import crypto
from bxcommon.utils.blockchain_utils.btc.btc_common_utils import btc_varint_to_int
from bxcommon.utils.blockchain_utils.btc.btc_object_hash import BtcObjectHash
from bxgateway.btc_constants import BTC_HDR_COMMON_OFF, BTC_BLOCK_HDR_SIZE, BTC_SHA_HASH_LEN, \
    BTC_COMPACT_BLOCK_SHORT_ID_LEN, BTC_SHORT_NONCE_SIZE, BTC_VARINT_MIN_SIZE
from bxgateway.messages.btc.btc_message import BtcMessage
from bxgateway.messages.btc.btc_message_type import BtcMessageType
from bxgateway.messages.btc.btc_messages_util import pack_int_to_btc_varint, get_next_tx_size, pack_block_header


class CompactBlockBtcMessage(BtcMessage, AbstractBlockMessage):
    """
    Message used to notify peers about new blocks.
    Blocks are sent using Compact Block message only if peer notified that would like to receive it instead of Inv message
    by sending Send Compact message after initial handshake.
    """

    MESSAGE_TYPE = BtcMessageType.COMPACT_BLOCK

    # pyre-fixme[9]: magic has type `int`; used as `None`.
    # pyre-fixme[9]: version has type `int`; used as `None`.
    # pyre-fixme[9]: prev_block has type `BtcObjectHash`; used as `None`.
    def __init__(self, magic: int = None, version: int = None, prev_block: BtcObjectHash = None,
                 # pyre-fixme[9]: merkle_root has type `BtcObjectHash`; used as `None`.
                 # pyre-fixme[9]: timestamp has type `int`; used as `None`.
                 # pyre-fixme[9]: bits has type `int`; used as `None`.
                 merkle_root: BtcObjectHash = None, timestamp: int = None, bits: int = None,
                 # pyre-fixme[9]: block_nonce has type `int`; used as `None`.
                 # pyre-fixme[9]: short_nonce has type `int`; used as `None`.
                 # pyre-fixme[9]: short_ids has type `List[memoryview]`; used as `None`.
                 block_nonce: int = None, short_nonce: int = None, short_ids: List[memoryview] = None,
                 # pyre-fixme[9]: prefilled_txns has type `List[memoryview]`; used
                 #  as `None`.
                 # pyre-fixme[9]: buf has type `bytearray`; used as `None`.
                 prefilled_txns: List[memoryview] = None, buf: bytearray = None):

        if buf is None:
            prefilled_tx_size = sum(len(tx) for tx in prefilled_txns)
            buf = bytearray(
                BTC_HDR_COMMON_OFF + BTC_BLOCK_HDR_SIZE + BTC_SHORT_NONCE_SIZE + 2 * BTC_VARINT_MIN_SIZE + BTC_COMPACT_BLOCK_SHORT_ID_LEN * len(
                    short_ids) + prefilled_tx_size)

            off = pack_block_header(buf, version, prev_block, merkle_root, timestamp, bits, block_nonce)

            struct.pack_into("<Q", buf, off, short_nonce)
            off += BTC_SHORT_NONCE_SIZE

            off += pack_int_to_btc_varint(len(short_ids), buf, off)

            for shortid in short_ids:
                buf[off:off + BTC_COMPACT_BLOCK_SHORT_ID_LEN] = shortid
                off += BTC_COMPACT_BLOCK_SHORT_ID_LEN

            off += pack_int_to_btc_varint(len(prefilled_txns), buf, off)

            for index, tx in prefilled_txns:
                # pyre-fixme[6]: Expected `Sized` for 1st param but got `int`.
                # pyre-fixme[6]: Expected `Union[typing.Iterable[int], bytes]` for
                #  2nd param but got `int`.
                buf[off:off + len(tx)] = tx
                # pyre-fixme[6]: Expected `Sized` for 1st param but got `int`.
                off += len(tx)

            self.buf = buf

            super(CompactBlockBtcMessage, self).__init__(magic, self.MESSAGE_TYPE, off - BTC_HDR_COMMON_OFF, buf)
        else:
            self.buf = buf
            self._memoryview = memoryview(buf)
            self._magic = self._command = self._payload_len = self._checksum = None
            self._payload = None

        self._version = self._prev_block = self._merkle_root = self._timestamp = None
        # pyre-fixme[8]: Attribute has type `int`; used as `None`.
        self._bits: int = None
        # pyre-fixme[8]: Attribute has type `int`; used as `None`.
        self._block_nonce: int = None
        # pyre-fixme[8]: Attribute has type `int`; used as `None`.
        self._short_nonce: int = None
        # pyre-fixme[8]: Attribute has type `int`; used as `None`.
        self._short_id_c: int = None
        self._short_nonce_buf = None
        # pyre-fixme[8]: Attribute has type `Dict[bytes, int]`; used as `None`.
        self._short_ids: Dict[bytes, int] = None
        # pyre-fixme[8]: Attribute has type `Dict[int, memoryview]`; used as `None`.
        self._pre_filled_transactions: Dict[int, memoryview] = None
        self._block_header = None
        # pyre-fixme[8]: Attribute has type `BtcObjectHash`; used as `None`.
        self._hash_val: BtcObjectHash = None
        self._pre_filled_tx_count = None

    def block_header(self) -> memoryview:
        if self._block_header is None:
            self._block_header = self._memoryview[BTC_HDR_COMMON_OFF:BTC_HDR_COMMON_OFF + BTC_BLOCK_HDR_SIZE]
        # pyre-fixme[7]: Expected `memoryview` but got `None`.
        return self._block_header

    def block_hash(self) -> BtcObjectHash:
        if self._hash_val is None:
            raw_hash = crypto.bitcoin_hash(self.block_header())
            self._hash_val = BtcObjectHash(buf=raw_hash, length=BTC_SHA_HASH_LEN)
        return self._hash_val

    def version(self) -> int:
        if self._version is None:
            off = BTC_HDR_COMMON_OFF
            self._version = struct.unpack_from("<I", self.buf, off)[0]
        # pyre-fixme[7]: Expected `int` but got `None`.
        return self._version

    def prev_block_hash(self) -> BtcObjectHash:
        if self._prev_block is None:
            off = BTC_HDR_COMMON_OFF + UL_INT_SIZE_IN_BYTES
            self._prev_block = BtcObjectHash(self.buf, off, BTC_SHA_HASH_LEN)
        # pyre-fixme[7]: Expected `BtcObjectHash` but got `None`.
        return self._prev_block

    def merkle_root(self) -> memoryview:
        if self._merkle_root is None:
            off = BTC_HDR_COMMON_OFF + UL_INT_SIZE_IN_BYTES + BTC_SHA_HASH_LEN
            self._merkle_root = self._memoryview[off:off + BTC_SHA_HASH_LEN]
        # pyre-fixme[7]: Expected `memoryview` but got `None`.
        return self._merkle_root

    def timestamp(self) -> int:
        if self._timestamp is None:
            off = BTC_HDR_COMMON_OFF + UL_INT_SIZE_IN_BYTES + 2 * BTC_SHA_HASH_LEN
            self._timestamp, self._bits, self._block_nonce, self._short_nonce = \
                struct.unpack_from("<IIIQ", self.buf, off)

        # pyre-fixme[7]: Expected `int` but got `None`.
        return self._timestamp

    def bits(self) -> int:
        if self._bits is None:
            self.timestamp()
        return self._bits

    def block_nonce(self) -> int:
        if self._block_nonce is None:
            self.timestamp()
        return self._block_nonce

    def short_nonce(self) -> int:
        if self._short_nonce is None:
            self.timestamp()
        return self._short_nonce

    def short_nonce_buf(self) -> memoryview:
        if self._short_nonce_buf is None:
            start = BTC_HDR_COMMON_OFF + BTC_BLOCK_HDR_SIZE
            self._short_nonce_buf = self._memoryview[start:start + BTC_SHORT_NONCE_SIZE]
        # pyre-fixme[7]: Expected `memoryview` but got `None`.
        return self._short_nonce_buf

    def short_ids(self) -> Dict[bytes, int]:
        if self._short_ids is None:
            off = BTC_HDR_COMMON_OFF + BTC_BLOCK_HDR_SIZE + BTC_SHORT_NONCE_SIZE
            self._short_id_c, size = btc_varint_to_int(self.buf, off)
            off += size
            self._short_ids = {}

            for idx in range(self._short_id_c):
                self._short_ids[bytes(self._memoryview[off:off + BTC_COMPACT_BLOCK_SHORT_ID_LEN])] = idx
                off += BTC_COMPACT_BLOCK_SHORT_ID_LEN

        return self._short_ids

    def pre_filled_transactions(self) -> Dict[int, memoryview]:
        if self._pre_filled_transactions is None:
            off = BTC_HDR_COMMON_OFF + BTC_BLOCK_HDR_SIZE + BTC_SHORT_NONCE_SIZE
            self._short_id_c, size = btc_varint_to_int(self.buf, off)
            off += size + BTC_COMPACT_BLOCK_SHORT_ID_LEN * self._short_id_c
            self._pre_filled_tx_count, size = btc_varint_to_int(self.buf, off)
            off += size

            self._pre_filled_transactions = {}

            index = -1
            # pyre-fixme[6]: Expected `int` for 1st param but got `None`.
            for _ in range(self._pre_filled_tx_count):
                diff, size = btc_varint_to_int(self.buf, off)
                off += size
                size = get_next_tx_size(self.buf, off)
                buf = self._memoryview[off:off + size]
                off += size

                index += diff + 1
                self._pre_filled_transactions[index] = buf

        return self._pre_filled_transactions

    def txns(self) -> List[str]:
        return [tx.hex() for tx in self.short_ids()]



import struct

from bxcommon.utils import convert
from bxcommon.utils.object_hash import Sha256Hash, PARTIAL_HASH_LENGTH
from bxcommon.utils.blockchain_utils.btc.btc_common_constants import BTC_SHA_HASH_LEN


class BtcObjectHash(Sha256Hash):
    def __init__(self, buf=None, offset=0, length=0, binary=None) -> None:

        from_binary = binary is not None and len(binary) == BTC_SHA_HASH_LEN
        from_buf = length == BTC_SHA_HASH_LEN and len(buf) >= offset + length

        if not (from_binary or from_buf):
            raise ValueError("Either binary or buf must contain data.")

        if buf is not None:
            if isinstance(buf, bytearray):
                local_binary = buf[offset:offset + length]
            else:  # In case this is a memoryview.
                local_binary = bytearray(buf[offset:offset + length])
            local_binary = local_binary[::-1]
        elif binary is not None:
            local_binary = binary
        else:
            raise ValueError("No data was passed")

        super(BtcObjectHash, self).__init__(local_binary)

        # This is where the big endian format will be stored
        self._buf = None
        self._hash = struct.unpack("<L", self.binary[-PARTIAL_HASH_LENGTH:])[0]
        self._full_str = None

    def full_string(self):
        if self._full_str is None:
            self._full_str = str(self.binary)
        return self._full_str

    def get_little_endian(self):
        return self.binary

    def get_big_endian(self):
        if self._buf is None:
            self._buf = self.binary[::-1]
        return self._buf

    def __repr__(self):
        return "BtcObjectHash<binary: {}>".format(convert.bytes_to_hex(self.binary))

    def __str__(self):
        return convert.bytes_to_hex(self.binary)


NULL_BTC_BLOCK_HASH = BtcObjectHash(binary=bytearray(BTC_SHA_HASH_LEN))



import struct

import blxr_rlp as rlp
from bxcommon.utils.blockchain_utils.eth import crypto_utils, eth_common_constants


class Frame(object):
    """
    When sending a packet over RLPx, the packet will be framed.
    The frame provides information about the size of the packet and the packet's
    source protocol. There are three slightly different frames, depending on whether
    or not the frame is delivering a multi-frame packet. A multi-frame packet is a
    packet which is split (aka chunked) into multiple frames because it's size is
    larger than the protocol window size (pws; see Multiplexing). When a packet is
    chunked into multiple frames, there is an implicit difference between the first
    frame and all subsequent frames.
    Thus, the three frame types are
    normal (single frame), chunked-0 (first frame of a multi-frame packet),
    and chunked-n (subsequent frames of a multi-frame packet).


    Single-frame packet:
    header || header-mac || frame || mac

    Multi-frame packet:
    header || header-mac || frame-0 ||
    [ header || header-mac || frame-n || ... || ]
    header || header-mac || frame-last || mac
    """

    header_data_serializer = rlp.sedes.List([rlp.sedes.big_endian_int] * 3, strict=False)

    def __init__(self, msg_type, payload, protocol_id=0,
                 sequence_id=None, total_payload_size=None, is_chunked=False):

        self._payload = payload if isinstance(payload, memoryview) else memoryview(payload)

        self._msg_type = msg_type
        self._sequence_id = sequence_id
        self._protocol_id = protocol_id
        self._is_chunked = is_chunked
        self._total_payload_size = total_payload_size

    def get_body_size(self, padded=False):
        """
        Returns size of frame body
        frame-size: 3-byte integer size of frame, big endian encoded (excludes padding)
        frame relates to body w/o padding w/o mac
        """

        l = len(self.get_encoded_msg_type()) + len(self._payload)
        if padded:
            l = crypto_utils.get_padded_len_16(l)
        return l

    def get_frame_size(self):
        """
        Returns total frame size
        Frame = header16 || mac16 || dataN + [padding] || mac16

        :return: frame size
        """

        return eth_common_constants.FRAME_HDR_DATA_LEN + eth_common_constants.FRAME_MAC_LEN + \
               self.get_body_size(padded=True) + eth_common_constants.FRAME_MAC_LEN

    def get_payload(self):
        """
        Returns frame payload
        :return: frame payload
        """

        return self._payload.tobytes()

    def get_protocol_id(self):
        """
        Return protocol id for frame
        :return: protocol id
        """

        return self._protocol_id

    def get_sequence_id(self):
        """
        Returns sequence id of chunked frame
        :return: sequence id
        """

        return self._sequence_id

    def is_chunked(self):
        """
        Returns flag indicating if frame is chunked
        :return: flag
        """

        return self._is_chunked

    def get_payload_len(self):
        """
        Returns length of frame payload
        :return: length
        """

        return len(self._payload)

    def get_total_payload_size(self):
        """
        Returns total payload size of chunked frames
        :return: total payload size of chunked frames
        """

        return self._total_payload_size

    def set_payload(self, payload):
        """
        Sets frame payload
        :param payload: payload
        """

        self._payload = memoryview(payload)

    def get_header(self):
        """
        Returns frame header

        header: frame-size || header-data || padding
        frame-size: 3-byte integer size of frame, big endian encoded
        header-data:
            normal: rlp.list(protocol-type[, sequence-id])
            chunked-0: rlp.list(protocol-type, sequence-id, total-packet-size)
            chunked-n: rlp.list(protocol-type, sequence-id)
            normal, chunked-n: rlp.list(protocol-type[, sequence-id])
            values:
                protocol-type: < 2**16
                sequence-id: < 2**16 (this value is optional for normal frames)
                total-packet-size: < 2**32
        padding: zero-fill to 16-byte boundary
        :return: frame header bytes
        """

        header_parts = [self._protocol_id]

        if self._is_chunked and self._sequence_id == 0:
            header_parts.append(self._sequence_id)
            header_parts.append(self._total_payload_size)
        elif self._sequence_id is not None:  # normal, chunked_n
            header_parts.append(self._sequence_id)

        header_data = rlp.encode(header_parts, sedes=self.header_data_serializer)

        # write body_size to header
        # frame-size: 3-byte integer size of frame, big endian encoded (excludes padding)
        # frame relates to body w/o padding w/o mac
        body_size = self.get_body_size()
        assert body_size < eth_common_constants.FRAME_MAX_BODY_SIZE
        header = struct.pack(">I", body_size)[1:] + header_data
        header = crypto_utils.right_0_pad_16(header)  # padding
        assert len(header) == eth_common_constants.FRAME_HDR_DATA_LEN
        return header

    def get_body(self):
        """
        Returns frame body
        frame:
            normal: rlp(packet-type) [|| rlp(packet-data)] || padding
            chunked-0: rlp(packet-type) || rlp(packet-data...)
            chunked-n: rlp(...packet-data) || padding
        padding: zero-fill to 16-byte boundary (only necessary for last frame)
        :return: frame body bytes
        """

        encoded_msg_type = self.get_encoded_msg_type()  # packet-type
        assert isinstance(self._payload, memoryview)
        return crypto_utils.right_0_pad_16(encoded_msg_type + self._payload.tobytes())

    def get_msg_type(self):
        """
        Returns message type
        :return: message type
        """

        return self._msg_type

    def get_encoded_msg_type(self):
        """
        Returns RLP-encoded message type
        :return: bytes of RLP-encoded message type
        """

        if self._sequence_id and self._sequence_id > 0:
            return b""

        return rlp.encode(self._msg_type, sedes=rlp.sedes.big_endian_int)  # unsigned byte



from typing import TypeVar, Iterator, Generic, Dict, Optional, Tuple

from bxcommon.utils.object_encoder import ObjectEncoder

TKeyRaw = TypeVar("TKeyRaw")
TKeyEncoded = TypeVar("TKeyEncoded")
TValueRaw = TypeVar("TValueRaw")
TValueEncoded = TypeVar("TValueEncoded")


class DefaultMapProxy(Generic[TKeyRaw, TKeyEncoded, TValueRaw, TValueEncoded]):

    def __init__(
            self,
            map_obj: Dict[TKeyRaw, TValueRaw],
            key_encoder: ObjectEncoder[TKeyRaw, TKeyEncoded],
            val_encoder: ObjectEncoder[TValueRaw, TValueEncoded]
    ):
        self.map_obj = map_obj
        self._key_encoder = key_encoder
        self._val_encoder = val_encoder
        if hasattr(self.map_obj, "__contains__"):
            self._contains = self._item_exists
        else:
            self._contains = self._item_exists_no_exception

    def __repr__(self):
        return self.map_obj.__repr__()

    def __str__(self) -> str:
        return self.map_obj.__str__()

    def __getitem__(self, key: TKeyEncoded) -> TValueEncoded:
        return self._val_encoder.encode(self.map_obj[self._key_encoder.decode(key)])

    def __len__(self) -> int:
        return len(self.map_obj)

    def __iter__(self) -> Iterator[TKeyEncoded]:
        for key in self.map_obj:
            yield self._key_encoder.encode(key)

    def __delitem__(self, key: TKeyEncoded):
        del self.map_obj[self._key_encoder.decode(key)]

    def __contains__(self, key: TKeyEncoded) -> bool:
        return self._contains(key)

    def pop(self, key: TKeyEncoded) -> TValueEncoded:
        if hasattr(self.map_obj, "pop"):
            return self._val_encoder.encode(self.map_obj.pop(self._key_encoder.decode(key)))
        else:
            val = self.__getitem__(key)
            self.__delitem__(key)
            return val

    def get(
            self, key: TKeyEncoded, default: Optional[TValueEncoded] = None
    ) -> Optional[TValueEncoded]:
        if self._contains(key):
            return self.__getitem__(key)
        else:
            return default

    def popitem(self) -> Tuple[TKeyEncoded, TValueEncoded]:
        key, val = next(iter(self.items()))
        self.__delitem__(key)
        return key, val

    def items(self) -> Iterator[Tuple[TKeyEncoded, TValueEncoded]]:
        for key, val in self.map_obj.items():
            yield self._key_encoder.encode(key), self._val_encoder.encode(val)

    def keys(self) -> Iterator[TKeyEncoded]:
        return self.__iter__()

    def values(self) -> Iterator[TValueEncoded]:
        for _, value in self.items():
            yield value

    def clear(self) -> None:
        self.map_obj.clear()

    def _item_exists_no_exception(self, key: TKeyEncoded) -> bool:
        try:
            self.__getitem__(key)
            return True
        except KeyError:
            return False

    def _item_exists(self, key: TKeyEncoded) -> bool:
        return self._key_encoder.decode(key) in self.map_obj
